/*
 * PDFlow -- filters.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "filters_priv.h"

static struct flt_interface *ifc [] = {
	&flt_ident_ifc, 
	&flt_hex_ifc, 
	&flt_a85_ifc, 
	&flt_flate_ifc, 
	&flt_lzw_ifc, 
	NULL
};

static struct v_option b_options [] = {
      	/*                                 Vers  sp  bs  sz */
	V_OPTION_BSPEED ("(none)",         1, 0, 10, 10,  5),
	V_OPTION_BSPEED ("FlateDecode",    1, 2,  4,  3, 10),
	V_END
};

static struct v_option a_options [] = {
      	/*                          Vers  sp  bs  sz text */
	V_OPTION ("(none)", 	    1, 0, 10, 10,  5, 1),
	V_OPTION ("ASCIIHexDecode", 1, 0,  5,  5,  0, 1),
	V_OPTION ("ASCII85Decode",  1, 0,  5,  5,  1, 1),
	V_END
};

static struct v_optable b_optable = {
	"comp_filters", b_options, ifc, sizeof (struct flt_interface *)
};

static struct v_optable a_optable = {
	"enc_filters", a_options, ifc, sizeof (struct flt_interface *)
};

static enum s_status find_filter (struct obj_bstr *name, 
				  struct flt_interface **f)
{
	int i;

	for (i = 0; ifc [i] != NULL; i++) {
		if (name -> size == strlen (ifc [i] -> name) &&
		    !memcmp (name -> s, ifc [i] -> name, name -> size))
			break;
	}

	if (ifc [i] == NULL)
		return s_make_indirect (S_CSC_NOT_SUPPORTED,
					"Filter %.*s is not supported",
					name -> size, name -> s);

	*f = ifc [i];

	return S_CSC_OK;
}

enum s_status flt_encode (struct obj_bstr *name, 
			  struct xref_db *rdb,
			  struct obj *params,
			  const struct obj_region *src, 
			  struct obj_region *dest)
{
	struct flt_interface *f;
	enum s_status res;

	res = find_filter (name, &f);
	if (res != S_CSC_OK)
		return res;

	return f -> encode (src, dest, rdb, params);
}

enum s_status flt_decode (struct obj_bstr *name, 
			  struct xref_db *rdb,
			  pdf_integer_t dl,
			  struct obj *params,
			  const struct obj_region *src, 
			  struct obj_region *dest)
{
	struct flt_interface *f;
	enum s_status res;

	res = find_filter (name, &f);
	if (res != S_CSC_OK)
		return res;

	return f -> decode (src, dest, rdb, dl, params);
}

enum s_status flt_add_selection (struct v_selection *sel)
{
	enum s_status res;

	res = v_add_selection (sel, &b_optable);
	if (res != S_CSC_OK)
		return res;

	res = v_add_selection (sel, &a_optable);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
