/*
 * PDFlow -- hex.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "filters_priv.h"

#define WS "\t\r\n\f "

static enum s_status encode (const struct obj_region *src, 
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     struct obj *parms);
static enum s_status decode (const struct obj_region *src, 
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     pdf_integer_t dl,
			     struct obj *parms);

struct flt_interface flt_hex_ifc = {
	"ASCIIHexDecode", encode, decode
};

static enum s_status encode (const struct obj_region *src, 
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     struct obj *parms)
{
	return S_CSC_NOT_SUPPORTED;
}

static enum s_status decode (const struct obj_region *src, 
			     struct obj_region *dest,
			     struct xref_db *rdb,
			     pdf_integer_t dl,
			     struct obj *parms)
{
	const unsigned char *p, *eos;
	char *q, *end, *buf;
	int c, d;

	eos = memchr (src -> begin, '>', src -> end - src -> begin);
	if (eos == NULL)
		return s_make_indirect (S_CSC_PARSE_ERROR,
					"Unterminated hex stream");

	c = eos - src -> begin;
	if (c & 1)
		c++;
	c >>= 1;
	buf = calloc (1, c ? c : 1);
	if (buf == NULL)
		return S_CSC_OUT_OF_MEMORY;

	q = end = buf;
	c = 0;
	for (p = src -> begin; p < eos; p++) {
		if (!*p || strchr (WS, *p) != NULL)
			continue;

		if (isdigit (*p))
			d = *p - '0';
		else if (isxdigit (*p))
			d = tolower (*p) - 'a' + 10;
		else
			return s_make_indirect (S_CSC_PARSE_ERROR,
						"Non-hex char '%c'", *p);

		if (c & 1) {
			*q |= d;
			q++;
		} else {
			*q |= d << 4;
			end++;
		}
		c++;
	}

	obj_init_malloc_region (dest, buf, end);

	return S_CSC_OK;
}

