/*
 * PDFlow -- xref_private.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef xref_private_h

#define xref_private_h

#include "refs.h"
#include "fmt/fields.h"

struct xref_cache_node {

	struct xref_cache_node *next;
	struct obj obj;
	void *ptr;
};

struct xref_cache {
	struct xref_cache *next;

	const char *id;

	struct ref_owner owner;

	void *closure;

	xref_cache_unref unref;

	struct xref_cache_node *list;
};

struct xref_subsection {

	struct xref_subsection *next;

	pdf_gen_t oid;

	pdf_integer_t entries;
};

struct xref_section {

	struct obj dict;

	pdf_integer_t trofs;

	struct xref_section *next;

	struct xref_subsection *head;

	struct xref_ifc *ifc;

	int secin;
};

struct xref_db {

	ref_t refs;

	pdf_integer_t id;

	pdf_integer_t entries;

	struct f_file *file, *scfile;

	struct xref_section *xrefs;

	struct ps_parser *parser;

	struct xref_cache *cache;
};

struct xref_ifc {
	enum s_status (*load) (struct xref_db *rdb,
			       struct xref_section **spp,
			       struct pl_state *state);

	enum s_status (*scratch) (struct xref_db *rdb,
				  struct xref_section **spp);

	enum s_status (*load_object) (struct xref_db *rdb,
				      struct xref_section *sp,
				      struct xref_subsection *ssp,
				      struct obj_reference *ref,
				      struct obj *obj);

	void (*cleanup) (struct xref_section *sp);

	void (*cleanup_subsection) (struct xref_subsection *ssp);

	enum s_status (*read_entry) (struct xref_db *rdb,
				     struct xref_section *sp,
				     struct xref_subsection *ssp,
				     pdf_integer_t oid,
				     struct obj_xref *xobj);

	enum s_status (*write_entry) (struct xref_db *rdb,
				      struct xref_section *sp,
				      struct xref_subsection *ssp,
				      pdf_integer_t oid,
				      struct obj_xref *xobj);
	
	enum s_status (*insert_subsection) (struct xref_db *rdb,
					    struct xref_section *sp,
					    struct xref_subsection **prev,
					    pdf_integer_t lo,
					    pdf_integer_t hi);

 	enum s_status (*sync_section) (struct xref_db *rdb,
				       struct xref_section *sp,
				       struct v_selection *sel);
};

void xref_add (struct xref_subsection **head, 
	       struct xref_subsection *ssp,
	       pdf_oid_t oid, pdf_integer_t entries);

void xref_default_cleanup (struct xref_section *sp);

enum s_status xref_serialize_trailer (struct xref_db *rdb,
				      struct v_selection *sel);

enum s_status xref_place_object (struct xref_db *rdb, struct obj *obj,
				 pdf_ofs_t ofs);
enum s_status xref_remove_object (struct xref_db *rdb, struct obj *obj);

extern struct xref_ifc xref10_ifc;
extern struct xref_ifc xref15_ifc;

#endif
