/*
 * PDFlow -- 2d.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "graph/2d.h"

#if 0
static void float__ref (struct td_float *tdf, struct ref_owner *p,
			const char *fname, int line)
{
	if (tdf != NULL)
		ref__up (&tdf -> refs, p, fname, line);
}
#define float_ref(tdp, p) float__ref (tdp, p, __FILE__, __LINE__)
#endif

static void float_unref (struct td_float *tdf, struct ref_owner *p)
{
	if (tdf == NULL || ref_down (&tdf -> refs, p))
		return;

	agc_unref (tdf -> agc, ref_rown (&tdf -> refs));
	
	ref_zero (&tdf -> refs);

	free (tdf);
}

static void set_space (struct td_graph_spaces *spaces, struct td_space *space,
		       enum gs_direction ldir, pdf_bool_t before)
{
	if (space == NULL || space -> type != TD_ST_GRAPHICAL)
		return;
	
	switch (ldir) {
	case GS_D_LEFT_TO_RIGHT:
		if (before)
			spaces -> left = space -> amount;
		else
			spaces -> right = space -> amount;
		break;

	case GS_D_RIGHT_TO_LEFT:
		if (before)
			spaces -> right = space -> amount;
		else
			spaces -> left = space -> amount;
		break;

	case GS_D_BOTTOM_TO_TOP:
		if (before)
			spaces -> bottom = space -> amount;
		else
			spaces -> top = space -> amount;
		break;

	case GS_D_TOP_TO_BOTTOM:
		if (before)
			spaces -> top = space -> amount;
		else
			spaces -> bottom = space -> amount;
		break;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}	
}

static void reserve (struct cobj_rect *avail, pdf_real_t size, 
		     enum gs_direction ldir)
{
	switch (ldir) {
	case GS_D_LEFT_TO_RIGHT:
		avail -> left += size;
		break;

	case GS_D_RIGHT_TO_LEFT:
		avail -> right -= size;
		break;

	case GS_D_BOTTOM_TO_TOP:
		avail -> bottom += size;
		break;

	case GS_D_TOP_TO_BOTTOM:
		avail -> top -= size;
		break;

	case GS_D_UNKNOWN:
	case GS_D__LAST__:
		assert (0);
	}
}

static enum s_status float_init (struct td_float *tdf, struct cobj_rect *rect,
				 enum gs_direction wdir, 
				 enum gs_direction ldir, 
				 struct td_space *space)
{
	enum s_status res;

	tdf -> wdir = wdir;
	tdf -> ldir = ldir;
	tdf -> avail = *rect;

	if (space != NULL)
		reserve (&tdf -> avail, space -> amount, ldir);

	set_space (&tdf -> spaces, space, ldir, 1);

	res = agc_float_new (&tdf -> agc, ref_rown (&tdf -> refs));
	if (res != S_CSC_OK)
		return res;
	
	return S_CSC_OK;
}

enum s_status float__new (struct td_float **tdfp, struct cobj_rect *rect,
			  enum gs_direction wdir, enum gs_direction ldir,
			  struct td_space *space,  struct ref_owner *p, 
			  const char *fname, int line)
{
	struct td_float *tdf;
	enum s_status res;

	tdf = malloc (sizeof (struct td_float));
	if (tdf == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memset (tdf, 0, sizeof (struct td_float));
	
	ref__init (&tdf -> refs, "td_float", p, fname, line);

	res = float_init (tdf, rect, wdir, ldir, space);
	if (res != S_CSC_OK) {
		float_unref (tdf, p);
		return res;
	}

	*tdfp = tdf;

	return S_CSC_OK;
}
#define float_new(tdf, rect, wdir, ldir, space, p) \
float__new (tdf, rect, wdir, ldir, space, p, __FILE__, __LINE__)

void td_pool_avail (struct td_pool *tdp, struct cobj_rect *rect)
{
	*rect = tdp -> floats != NULL ? 
		tdp -> floats -> avail : tdp -> avail;
}

void td_pool_line_avail (struct td_pool *tdp, struct cobj_rect *rect)
{
	td_pool_avail (tdp, rect);
	if (tdp -> floats != NULL) {
		if (tdp -> floats -> wdir == GS_D_LEFT_TO_RIGHT ||
		    tdp -> floats -> wdir == GS_D_RIGHT_TO_LEFT) {
			rect -> top = tdp -> total.top;
			rect -> bottom = tdp -> total.bottom;
		} else {
			rect -> left = tdp -> total.left;
			rect -> right = tdp -> total.right;
		}
	}
}

static void get_vectors (enum gs_direction dir, struct cobj_rect *rect,
			 struct gs_point *p0, struct gs_point *inc,
			 pdf_bool_t *postinc)
{
	switch (dir) {
	case GS_D_LEFT_TO_RIGHT:
		if (postinc != NULL)
			*postinc = 1;
		if (p0 != NULL)
			p0 -> x = rect -> left;
		inc -> x = +1;
		inc -> y = 0;
		break;

	case GS_D_RIGHT_TO_LEFT:
		if (postinc != NULL)
			*postinc = 0;
		if (p0 != NULL)
			p0 -> x = rect -> right;
		inc -> x = -1;
		inc -> y = 0;
		break;

	case GS_D_TOP_TO_BOTTOM:
		if (postinc != NULL)
			*postinc = 0;
		if (p0 != NULL)
			p0 -> y = rect -> top;
		inc -> x = 0;
		inc -> y = -1;
		break;

	case GS_D_BOTTOM_TO_TOP:
		if (postinc != NULL)
			*postinc = 1;
		if (p0 != NULL)
			p0 -> x = rect -> bottom;
		inc -> x = 0;
		inc -> y = +1;
		break;

	default:
		assert (0);
	}
}

static void move_point_r (struct gs_point *point, struct gs_point *vec,
			  struct cobj_rect *rect)
{
	point -> x += vec -> x * cobj_rect_width (rect);
	point -> y += vec -> y * cobj_rect_height (rect);
}

static void move_point_d (struct gs_point *point, struct gs_point *vec,
			  pdf_real_t r)
{
	point -> x += vec -> x * r;
	point -> y += vec -> y * r;
}

enum s_status td_float_bol (struct td_line_status *tdl, struct td_float *tdf, 
			    pdf_real_t ascent, pdf_real_t descent)
{
	tdl -> tdf = tdf;

	get_vectors (tdf -> wdir, &tdf -> avail, &tdl -> point,
		     &tdl -> wvec, &tdl -> wpi);
	get_vectors (tdf -> ldir, &tdf -> avail, &tdl -> point,
		     &tdl -> lvec, &tdl -> lpi);

	reserve (&tdf -> avail, ascent + descent, tdf -> ldir);
	if (!tdl -> lpi)
		move_point_d (&tdl -> point, &tdl -> lvec, ascent);

	return S_CSC_OK;
}

enum s_status td_float_write (struct td_line_status *tdl, 
			      struct agc_context *agc,
			      struct cobj_rect *rect, struct gs_point *origin)
{
	struct gs_point tpoint;
	enum s_status res;

	if (!tdl -> wpi)
		move_point_r (&tdl -> point, &tdl -> wvec, rect);

	tpoint.x = tdl -> point.x - origin -> x;
	tpoint.y = tdl -> point.y - origin -> y;

	res = agc_apply (agc, tdl -> tdf -> agc, &tpoint);
	if (res != S_CSC_OK)
		return res;

	if (tdl -> wpi)
		move_point_r (&tdl -> point, &tdl -> wvec, rect);

	return S_CSC_OK;
}

enum s_status td_float_space (struct td_line_status *tdl, pdf_real_t amount)
{
	move_point_d (&tdl -> point, &tdl -> wvec, amount);

	return S_CSC_OK;
}

enum s_status td_float_eol (struct td_line_status *tdl)
{
	return S_CSC_OK;
}

static enum s_status pool_init (struct td_pool *tdp, struct cobj_rect *rect)
{
	tdp -> avail = tdp -> total = *rect;

	return S_CSC_OK;
}

void td_pool__ref (struct td_pool *tdp, struct ref_owner *p,
		   const char *fname, int line)
{
	if (tdp != NULL)
		ref__up (&tdp -> refs, p, fname, line);
}

static void destroy_floats (struct td_pool *tdp)
{
	struct td_float *tdf;

	while (tdp -> floats != NULL) {
		tdf = tdp -> floats;
		tdp -> floats = tdf -> next;
		float_unref (tdf, ref_rown (&tdp -> refs));
	}
}

void td_pool_unref (struct td_pool *tdp, struct ref_owner *p)
{
	if (tdp == NULL || ref_down (&tdp -> refs, p))
		return;

	destroy_floats (tdp);

	ref_zero (&tdp -> refs);

	free (tdp);
}

enum s_status td_pool__new (struct td_pool **tdpp, struct cobj_rect *rect,
			    struct ref_owner *p, const char *fname, int line)
{
	struct td_pool *tdp;
	enum s_status res;

	tdp = malloc (sizeof (struct td_pool));
	if (tdp == NULL)
		return S_CSC_OUT_OF_MEMORY;

	memset (tdp, 0, sizeof (struct td_pool));

	ref__init (&tdp -> refs, "td_pool", p, fname, line);
	
	res = pool_init (tdp, rect);
	if (res != S_CSC_OK) {
		td_pool_unref (tdp, p);
		return res;
	}

	*tdpp = tdp;

	return S_CSC_OK;
}

enum s_status td_float_alloc (struct td_float **tdfp, struct td_pool *tdp,
			      enum gs_direction wdir, enum gs_direction ldir,
			      struct td_space *space)
{
	struct td_float *tdf;
	enum s_status res;

	res = float_new (&tdf, &tdp -> avail, wdir, ldir, 
			 space, ref_rown (&tdp -> refs));
	if (res != S_CSC_OK)
		return res;

	tdf -> next = tdp -> floats;
	tdp -> floats = tdf;

	*tdfp = tdf;

	return S_CSC_OK;
}

void td_float_done (struct td_pool *tdp, struct td_float *tdf,
		    struct td_space *space)
{
	assert (tdf == tdp -> floats);
	if (space != NULL)
		set_space (&tdf -> spaces, space, tdf -> ldir, 0);
	tdf -> rect = tdp -> avail;
	cobj_rect_sub (&tdf -> rect, &tdf -> avail);
	tdp -> avail = tdf -> avail;
}

static void distribute_graph_spaces (struct td_pool *tdp, struct gs_point *req)
{
	struct td_float *tdf;
	pdf_real_t hspace, vspace;
	pdf_real_t hreq, vreq;
	pdf_real_t hunit, vunit;

	if (cobj_rect_width (&tdp -> total) == 
	    cobj_rect_width (&tdp -> avail))
		vspace = cobj_rect_height (&tdp -> avail);
	else
		vspace = 0;

	if (cobj_rect_height (&tdp -> total) == 
	    cobj_rect_height (&tdp -> avail))
		hspace = cobj_rect_width (&tdp -> avail);
	else
		hspace = 0;

	if (!hspace && !vspace) {
		req -> x = req -> y = 0;
		return;
	}

	hreq = vreq = 0;
	for (tdf = tdp -> floats; tdf != NULL; tdf = tdf -> next) {
		hreq += tdf -> spaces.left + tdf -> spaces.right;
		vreq += tdf -> spaces.top + tdf -> spaces.bottom;
	}

	if (!hreq && !vreq) {
		req -> x = req -> y = 0;
		return;
	}

	if (hreq) {
		hunit = hspace / hreq;
		req -> x = cobj_rect_width (&tdp -> avail);
	} else {
		hunit = 0;
		req -> x = 0;
	}

	if (vreq) {
		vunit = vspace / vreq;
		req -> y = cobj_rect_height (&tdp -> avail);
	} else {
		vunit = 0;
		req -> y = 0;
	}

	for (tdf = tdp -> floats; tdf != NULL; tdf = tdf -> next) {
		tdf -> spaces.left *= hunit;
		tdf -> spaces.right *= hunit;
		tdf -> spaces.bottom *= vunit;
		tdf -> spaces.top *= vunit;
	}
}

static void relocate_rects (struct td_pool *tdp, struct gs_point *req)
{
	struct td_float *tdf, *p;

	for (tdf = tdp -> floats; tdf != NULL; tdf = tdf -> next) {
		tdf -> origin.x = tdf -> spaces.left - req -> x;
		tdf -> origin.y = tdf -> spaces.bottom - req -> y;

		for (p = tdp -> floats; p != NULL; p = p -> next) {
			if (p -> rect.left < tdf -> rect.left)
				tdf -> origin.x += 
					p -> spaces.left + 
					p -> spaces.right;
			if (p -> rect.bottom < tdf -> rect.bottom)
				tdf -> origin.y += 
					p -> spaces.bottom + 
					p -> spaces.top;
		}
	}
}

static enum s_status draw (struct td_pool *tdp, struct agc_context *agc)
{
	struct td_float *tdf;
	enum s_status res;

	for (tdf = tdp -> floats; tdf != NULL; tdf = tdf -> next) {
		res = agc_apply (tdf -> agc, agc, &tdf -> origin);
		if (res != S_CSC_OK)
			return res;
	}

	return S_CSC_OK;
}

enum s_status td_pool_finalize (struct td_pool *tdp, struct agc_context **agcp,
				struct ref_owner *p)
{
	struct agc_context *agc;
	struct gs_point req;
	enum s_status res;

	res = agc_float_new (&agc, p);
	if (res != S_CSC_OK)
		return res;

	distribute_graph_spaces (tdp, &req);
	relocate_rects (tdp, &req);
	res = draw (tdp, agc);
	destroy_floats (tdp);
	tdp -> avail = tdp -> total;
	if (res != S_CSC_OK) {
		agc_unref (agc, p);
		return res;
	}

	*agcp = agc;

	return S_CSC_OK;
}
