/*
 * PDFlow -- baseline.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <assert.h>
#include <string.h>

#include "grid/baseline.h"

struct bl_closure {
	struct args_heur *heur;
};


static enum s_status do_fix_baselines (struct gr_grid *grid,
				       struct args_heur *heur)
{
	struct heur_bl_chooser hbl;
	struct gr_line *line;
	struct gr_cell *cell;
	pdf_bool_t has_origin, horiz;


	switch (grid -> ifc -> logical) {
	case GR_L_WORD:
		a_get_origin (grid -> d.area, grid -> wdir, heur, 
			      &grid -> origin, &has_origin);
		break;

	case GR_L_LINE:
		line = grid -> d.lines;

		heur_bl_init (&hbl, heur, grid -> wdir);
		for (cell = line -> cells; cell != NULL; cell = cell -> next)
			heur_bl_add (&hbl, &cell -> grid -> origin);

		heur_bl_choose (&hbl, &grid -> g.rect, 
				&grid -> origin, &has_origin);

		horiz = grid -> wdir == GS_D_TOP_TO_BOTTOM ||
			grid -> wdir == GS_D_BOTTOM_TO_TOP;

		for (cell = line -> cells; cell != NULL; cell = cell -> next)
			if (grid -> lines == 1) {
				if (horiz)
					cell -> grid -> origin.x = 
						grid -> origin.x;
				else
					cell -> grid -> origin.y = 
						grid -> origin.y;
			}
		break;

	default:
		break;
	}

	return S_CSC_OK;
}

static enum s_status fix_baselines (struct gr_grid *grid, 
				    void *closure)
{
	struct bl_closure *blc;
	enum s_status res;

	blc = closure;

	res = gr_traverse (grid, fix_baselines, blc);
	if (res != S_CSC_OK)
		return res;

	res = do_fix_baselines (grid, blc -> heur);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status gr_bl_fix_baselines (struct gr_grid *grid, 
				   struct args_heur *heur)
{
	struct bl_closure blc;
	enum s_status res;

	blc.heur = heur;
	res = fix_baselines (grid, &blc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
