/*
 * PDFlow -- qi.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "grid/thread.h"
#include "grid/qi.h"

struct classify_closure {
	struct args_heur *heur;

	struct gr_grid *last_cell;
};

static void close_rlen (struct gr_grid_fill *gfill, struct args_heur *heur)
{
	struct gr_cell *cell;
	int i;

	if (gfill -> run_len >= heur -> fill_thresh) {
		cell = gfill -> first;
		for (i = 0; i < gfill -> run_len; i++) {
			cell -> grid -> fill.ridx = i;
			cell = cell -> next;
		}
	} else if (gfill -> run_len)
		gfill -> first = NULL;

	gfill -> run_len = 0;
}

static void classify_fill (struct gr_grid_fill *gfill, struct gr_cell *cell,
			   int is_fill, struct args_heur *heur)
{
	/* Algorithm may be disabled */
	if (heur -> fill_thresh <= 0)
		return;

	if (is_fill) {
		if (gfill -> run_len)
			gfill -> run_len++;
		else if (gfill -> first == NULL) {
			gfill -> run_len = 1;
			gfill -> first = cell;
		}
	} else 
		close_rlen (gfill, heur);
}

static enum s_status classify (struct gr_grid *grid, 
			       struct gr_th_info *info,
			       void *closure)
{
	struct classify_closure *cc;
	struct a_info ainfo;

	cc = closure;

	if (info -> parent == NULL ||
	    info -> parent -> ifc -> logical != GR_L_LINE)
		return S_CSC_OK;

	if (grid -> ifc -> logical == GR_L_WORD) {
		a_query_info (grid -> d.area, info -> parent -> wdir, &ainfo);
		classify_fill (&info -> parent -> fill, 
			       info -> cell, ainfo.fill, cc -> heur);
		if (info -> last) {
			if (ainfo.hyphenized) {
				grid -> hyphen = ainfo.hyphen;
				ainfo.hyphen -> code = GR_AGCC_HYPHEN;
			}
			grid -> hyphen = 
				ainfo.hyphenized ? ainfo.hyphen : NULL;

			/* Close the fill algorithm */
			close_rlen (&info -> parent -> fill, cc -> heur);
		}
	}

	return S_CSC_OK;
}

enum s_status gr_qi_classify_lines (struct gr_grid *grid,
				    struct args_heur *heur)
{
	struct classify_closure cc;
	enum s_status res;

	cc.heur = heur;

	res = gr_th_traverse (grid, heur, classify, &cc);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
