/*
 * PDFlow -- sig.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

/*
 * The algorithm is just a skeleton: we only look at the average
 * line size.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>

#include "grid/sig.h"

void gr_sig_mksig (struct gr_sig_signature *sig, struct gr_grid *grid)
{
	struct gr_line *line;
	struct gr_cell *cell;
	pdf_real_t avw, avh;
	unsigned n;

	assert (grid -> ifc -> logical == GR_L_LINE);

	line = grid -> d.lines;

	avh = avw = n = 0;
	for (cell = line -> cells; cell != NULL; cell = cell -> next) {
		avh += cobj_rect_height (&cell -> grid -> g.rect);
		avw += cobj_rect_width (&cell -> grid -> g.rect);
		n++;
	}

	if (n)
		sig -> size = 
			grid -> wdir == GS_D_LEFT_TO_RIGHT || 
			grid -> wdir == GS_D_RIGHT_TO_LEFT ?
			avh / n : avw / n;
	else
		sig -> size = 0;
			
}

pdf_real_t gr_sig_affinity (struct gr_sig_signature *sig1,
			    struct gr_sig_signature *sig2,
			    struct args_heur *heur)
{
	if (!sig1 -> size)
		return sig2 -> size ? 0 : 1;
	else if (!sig2 -> size)
		return 0;
	else
		return sig1 -> size > sig2 -> size ? 
			sig2 -> size / sig1 -> size : 
			sig1 -> size / sig2 -> size;
}


pdf_real_t gr_sig_can_merge (struct gr_sig_signature *sig1,
			     struct gr_sig_signature *sig2,
			     struct args_heur *heur)
{
	return gr_sig_affinity (sig1, sig2, heur) >=
		heur -> merge_line_affinity;
}
