/*
 * PDFlow -- wspace.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "grid/wspace.h"
#include "grid/thread.h"

static enum s_status first_pass (struct gr_grid *grid, 
				 struct gr_th_info *info,
				 void *closure)
{
	if (info -> wspace.type != TD_ST_GRAPHICAL &&
	    info -> wspace.amount >= 0)
		heur_hls_add (&grid -> hls, info -> wspace.amount);

	return S_CSC_OK;
}

static enum s_status second_pass (struct gr_grid *grid, 
				  struct gr_th_info *info,
				  void *closure)
{
	if (info -> wspace.type == TD_ST_GRAPHICAL)
		grid -> wspace = info -> wspace;
	else if (info -> wspace.amount >= 0) {
		grid -> wspace.type = TD_ST_ORDINARY;
		heur_hls_choose (&grid -> hls, info -> wspace.amount, 
				 &grid -> wspace.amount);
	} else {
		grid -> wspace.type = TD_ST_ORDINARY;
		grid -> wspace.amount = 1;
	}

	return S_CSC_OK;
}

enum s_status gr_ws_fix_spaces (struct gr_grid *grid,
				struct args_heur *heur)
{
	enum s_status res;

	res = gr_th_traverse (grid, heur, first_pass, NULL);
	if (res != S_CSC_OK)
		return res;

	res = gr_th_traverse (grid, heur, second_pass, NULL);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

