/*
 * PDFlow -- args.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef args_h

#define args_h

#include "version.h"

#define ARGS_MAX_MARGINS 30

enum args_app {

	ARGS_A_HIGHLIGHT,
	ARGS_A_FLOAT,
	ARGS_A_RESIZE,
	ARGS_A_DETECT

};

enum args_margin_type {
	ARGS_LD_NORTH,
	ARGS_LD_EAST,
	ARGS_LD_WEST,
	ARGS_LD_SOUTH
};

struct args_margin {
	
	enum args_margin_type type;

	pdf_real_t value;

};

struct args_rect {

	pdf_real_t left, right, top, bottom;

};

struct args_margins {

	pdf_bool_t skip_m, skip_w;

	pdf_integer_t margins, max_margins, quorum;

	pdf_real_t weight;

	struct args_rect clip;

	struct args_margin margin [ARGS_MAX_MARGINS];

};

struct args_heur {

	pdf_real_t column_glue;

	pdf_real_t word_glue, glyph_glue;
	
	pdf_real_t indent_tolerance, max_indent;
	
	pdf_real_t baseline_tolerance;

	pdf_real_t hf_tolerance, hf_percentage, hf_gap;

	pdf_real_t header_thresh, footer_thresh;

	pdf_real_t par_decadence, par_break;

	pdf_real_t graph_space, w_graph_space, w_tolerance;

	pdf_real_t weight_tolerance;

	pdf_integer_t fill_thresh;

	pdf_real_t merge_line_affinity;
};

struct args_output {

	pdf_real_t width, height;

	pdf_real_t text_ratio, space_ratio;

	pdf_real_t graph_ratio, graph_bad_slope;

	pdf_real_t wspace_width;

	pdf_real_t graph_space_el;

	const char *hlspec;
};

struct args_range {
	pdf_integer_t lo, hi;
};

struct args_ranges {

	struct args_range doc;
	
	struct args_range detect;

};

struct args_debug {
	char *report;
	FILE *rfp;
};

struct args_file {
	char *src, *dest;
};

struct args {

	enum args_app app;
	
	struct v_pref prefs;

	struct args_heur heur;

	struct args_margins amys;

	struct args_output aout;

	struct args_ranges ranges;

	struct args_debug debug;

	int files;
	struct args_file *file;
};

enum s_status args_init (struct args *args, int argc, char *argv []);
void args_destroy (struct args *args);

#endif
