/*
 * PDFlow -- ctab.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef ctab_h

#define ctab_h

#include <stdint.h>
#include <stdlib.h>

#include "status.h"
#include "fmt/obj.h"

enum ct_encoding_type {
	CT_ET_STANDARD,
	CT_ET_MAC_ROMAN,
	CT_ET_WIN_ANSI,
	CT_ET_PDF_DOC,
	CT_ET_MAC_EXPERT,
	CT_ET_MAC_OS_ROMAN,

	CT_ET__LAST
};

enum ct_t0_encoding_type {
	CT_ET0_IDENTITY_H,
	CT_ET0_IDENTITY_V
};

struct ct_encoding {
        char *name;
	uint32_t unicode;
};

struct ct_encoding_table {
	unsigned namelen;
	struct ct_encoding e [256];
};

int ct_utf8_len (uint32_t c);

char *ct_utf8_put (char *p, uint32_t c);

enum s_status ct_utf16_to_utf8 (const void *utf16, size_t size,
				char **utf8);

enum s_status ct_pdfdoc_to_utf8 (const void *pdfdoc, char **utf8);

enum s_status ct_fill_encoding (struct ct_encoding_table *e, 
				enum ct_encoding_type et);
void ct_destroy_encoding (struct ct_encoding_table *e);

uint32_t ct_utf16_to_unicode (const void *utf16, size_t size);
uint32_t ct_name_to_unicode (const char *name);
uint32_t ct_name_to_unicode_s (const struct obj_bstr *name);
uint32_t ct_char_to_unicode (int code, enum ct_encoding_type et);
const char *ct_char_to_name (int code, enum ct_encoding_type et);

pdf_bool_t ct_is_hyphen (uint32_t unicode);
pdf_bool_t ct_is_fill (uint32_t unicode);

#endif
