/*
 * PDFlow -- fields.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef fields_h

#define fields_h

#include <stdarg.h>


#include "fmt/obj.h"
struct xref_obj_array {
	size_t n;
	struct obj *data;
};

#include "fmt/xref.h"
#include "version.h"

enum xref_op {
 	XREF_OP_PARSE,
	XREF_OP_SERIALIZE,
	XREF_OP_DESTROY
};

enum xref_deserialize {
	XREF_D_UNWRAP,
	XREF_D_OBJ,
	XREF_D_PSTABLE,
	XREF_D_ARRAY,
	XREF_D_NEST,
	XREF_D_DYNAMIC,
	XREF_D_REGION
};

enum xref_serialize {

	XREF_S_INLINE,
	XREF_S_REFERENCE
};

enum xref_changed {
	XREF_C_CHANGED,
	XREF_C_UNCHANGED,
	XREF_C_UNKNOWN
};

#define XREF_BIND_SIZE 20
struct xref_bind {
	int size;

	struct {
		const char *name;
		void *p;
	} e [XREF_BIND_SIZE];
};

struct xref_field;
typedef enum s_status (*xref_validator_t) (void *data);
typedef enum s_status (*xref_dyn_ifc_t) (enum xref_op op, 
					 struct xref_field *fd, void *data,
					 int *stop);

extern const struct xref_state xref_mandatory_tag, xref_optional_tag;

struct xref_field {
	const char *name;
	void *field;
	enum obj_type ot;
	enum xref_deserialize des;
	enum xref_serialize ser;
	int major, minor;
	const struct xref_state *state;
	void *data;
	struct obj defval;
	int simplify;
	union {
		/* XREF_D_ARRAY or XREF_D_PSTABLE */
		struct xref_pstable *pstable;
		/* XREF_D_DYNAMIC */
		xref_dyn_ifc_t dyn;
		/* XREF_D_NEST */
		struct xref_field_desc *fdesc;
	} u;
};

struct xref_field_desc {
	xref_validator_t val;
	struct xref_field fd [];
};

typedef enum s_status (*xref_parser_t) (struct xref_db *rdb, 
					struct xref_bind *robj,
					struct ref_owner *p,
					struct obj *src,
					struct xref_field *fd,
					const char *fname, int line);
typedef enum s_status (*xref_serializer_t) (struct xref_db *rdb, 
					    struct xref_bind *robj,
					    struct v_selection *sel,
					    struct obj *dest,
					    enum xref_changed *changed,
					    struct xref_field *fd);
typedef void (*xref_funref_t) (struct ref_owner *p,
			       struct xref_field *fd);

struct xref_pstable {
	size_t size;
	xref_parser_t parser;
	xref_serializer_t serializer;
	xref_funref_t funref;
};

#ifdef DEBUG
int *xref_p_cast (xref_parser_t f);
xref_serializer_t xref_s_cast (xref_serializer_t f);
#else
#define xref_p_cast(f) ((int *) (f))
#define xref_s_cast(f) ((f))
#endif

struct xref_mapped_array {
	size_t n;
	void *data;
};

void xref_bind_init (struct xref_bind *robj);
void xref_bind_data (struct xref_bind *robj, const char *name, void *p);
void *xref_bind_get (struct xref_bind *robj, const char *name);

void xref_init_field (struct xref_field *fd, void *field, void *data);

enum s_status xref_dict_resolve (struct xref_db *rdb, struct obj *dict,
				 const char *key, enum obj_type ot, 
				 struct obj *obj);

enum s_status xref__fields_parse (struct xref_db *rdb, struct xref_bind *robj,
				  struct ref_owner *p, struct obj *obj, 
				  struct xref_field *fd, void *base,
				  const char *fname, int line);
#define xref_fields_parse(rdb, robj, p, obj, fd, base) \
xref__fields_parse (rdb, robj, p, obj, fd, base, __FILE__, __LINE__)

enum s_status xref__fields_replace (struct xref_db *rdb, 
				    struct xref_bind *robj,
				    struct ref_owner *p, struct obj *obj, 
				    struct xref_field *fd, void *base,
				    const char *fname, int line);
#define xref_fields_replace(rdb, robj, p, obj, fd, base) \
xref__fields_replace (rdb, robj, p, obj, fd, base, __FILE__, __LINE__)

enum s_status xref_fields_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, 
			       struct obj *sobj, enum xref_changed *changed,
			       struct xref_field *fd, void *base);

enum s_status xref__array_map_ps (struct xref_db *rdb, struct xref_bind *robj,
				  struct ref_owner *p, struct obj *src, 
				  void *mapped_array, const char *label, 
				  struct xref_pstable *pst, void *data,
				  const char *fname, int line);
#define xref_array_map_ps(rdb, ro, p, src, ma, label, pst, data)\
xref__array_map_ps (rdb, ro, p, src, ma, label, pst, data, __FILE__, __LINE__)

enum s_status xref_array_unmap (struct xref_db *rdb, struct xref_bind *robj,
				struct v_selection *sel, void *mapped_array,
				struct obj *dest, size_t elsize, int simplify,
				xref_serializer_t ser, void *pdata);

enum s_status xref_array_unmap_ps (struct xref_db *rdb, struct xref_bind *robj,
				   struct v_selection *sel, void *mapped_array,
				   struct obj *dest, int simplify,
				   struct xref_pstable *pst, void *data);

void xref_array_map_rel (void *mapped_array, struct ref_owner *p,
			 size_t elsize, xref_funref_t funref, void *pdata);
void xref_array_map_rel_ps (void *mapped_array, struct ref_owner *p,
			    struct xref_pstable *pst, void *data);

void xref_fields_destroy (struct xref_field *fd, void *base,
			  struct ref_owner *p);

enum s_status xref_dict_is (struct xref_db *rdb, struct obj *obj,
			    const char *type, pdf_bool_t *res);

extern struct xref_pstable xref_tag_pstable;

extern struct xref_pstable xref_bool_pstable;
extern struct xref_pstable xref_int_pstable, xref_real_pstable;
extern struct xref_pstable xref_name_pstable, xref_dictionary_pstable;
extern struct xref_pstable xref_stream_pstable, xref_obj_pstable;

struct xref_bool_array {
	size_t n;
	pdf_bool_t *data;
};

struct xref_int_array {
	size_t n;
	pdf_integer_t *data;
};

struct xref_real_array {
	size_t n;
	pdf_real_t *data;
};

#define XREF_A_SIZE(a) 	   ((a) -> n == -1 ? 0 : (a) -> n)
#define XREF_A_IS_NULL(a)  ((a) -> n == -1)
#define XREF_A_SET_NULL(a) do { (a) -> n = -1; (a) -> data = NULL; } while (0)


#endif

