/*
 * PDFlow -- fields.tab.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#undef XREF__F
#undef XREF__D
#undef XREF__T

#undef XREF_DESC_BEGIN
#undef XREF_DESC_END

#ifdef XREF_IMPLEMENTATION

#ifndef XREF_TYPE
#define XREF_TYPE XREF_IMPLEMENTATION
#endif

#ifndef XREF_M
#define XREF_M &xref_mandatory_tag
#endif

#ifndef XREF_O
#define XREF_O &xref_optional_tag
#endif

#define XREF__F(name, ftype, field, ot, des, ser, major, minor, mand, 	\
	        uf, ufd, data, dot, dtag, dval, simplify)             	\
{ name, &((XREF_TYPE *) 0) -> field, ot, des, ser, major, minor,   	\
  mand, data, { type : dot, val : { dtag : dval } }, simplify, { uf: ufd} },
#define XREF__T(name, type, field, ot, des, ser, major, minor, mand, 	\
	        uf, ufd, data, defdesc, simplify)                    	\
{ name, NULL, ot, des, ser, major, minor,   				\
  mand, data, defdesc, simplify, { uf: ufd} },
#define XREF__D(name, ot, ser, des, major, minor, uf, ufd, data, simplify)\
static struct xref_field name = 					\
{ #name, NULL, ot, des, ser, major, minor, 				\
  XREF_M, data, { }, simplify, { uf: ufd } };
#else
#define XREF__F(name, type, field, ot, des, ser, major, minor, mand, 	\
	        uf, ufd, data, dot, dtype, dval, simplify)            	\
	type field;
#define XREF__T(name, type, field, ot, des, ser, major, minor, mand, 	\
	        uf, ufd, data, defdesc, simplify)
#define XREF__D(name, ot, ser, des, major, minor, uf, ufd, data, simplify)
#endif

#define XREF__FND(name, type, field, ot, des, ser, major, minor, mand, \
                  dot, dtag, dval)\
XREF__F (name, type, field, ot, des, ser, major, minor, mand, \
         pstable, NULL, NULL, dot, dtag, dval, 0)

#define XREF__SCALAR(name, type, field, ot, ufield, defv, major, minor, mand)\
XREF__FND (name, type, field, ot, XREF_D_UNWRAP, XREF_S_INLINE, \
           major, minor, mand, ot, ufield, defv)

#define XREF_TAG(name, value, major, minor, mand) \
XREF__T (name, <none>, <none>, OT_NAME, XREF_D_PSTABLE, XREF_S_INLINE, \
         major, minor, mand, pstable, &xref_tag_pstable, NULL, \
         { val: { string: { s: value } } }, 0)

#define XREF_BOOL(name, field, def, major, minor, mand) \
XREF__SCALAR (name, pdf_bool_t, field, OT_BOOLEAN, bool, def, \
              major, minor, mand)
#define XREF_INTEGER(name, field, def, major, minor, mand) \
XREF__SCALAR (name, pdf_integer_t, field, OT_INTEGER, integer, def, \
              major, minor, mand)
#define XREF_REAL(name, field, def, major, minor, mand) \
XREF__SCALAR (name, pdf_real_t, field, OT_REAL, real,  def, \
	      major, minor, mand)

#define XREF_OBJ(name, field, ot, major, minor, mand) \
XREF__FND (name, struct obj, field, ot, XREF_D_OBJ, XREF_S_REFERENCE, \
           major, minor, mand, OT_NULL, integer, 0)
#define XREF_INLINE(name, field, ot, major, minor, mand) \
XREF__FND (name, struct obj, field, ot, XREF_D_OBJ, XREF_S_INLINE, \
           major, minor, mand, OT_NULL, integer, 0)

#define XREF__FUNCTION_DATA(name, type, field, ot, pst, data, \
                            major, minor, mand, ser, simplify) \
XREF__F (name, type, field, ot, XREF_D_PSTABLE, ser, \
         major, minor, mand, pstable, pst, data, OT_NULL, integer, 0, simplify)
#define XREF__ARRAY(name, type, field, ot, pst, data, \
                    major, minor, mand, ser, simplify) \
XREF__F (name, type, field, ot, XREF_D_ARRAY, ser, \
         major, minor, mand, pstable, pst, data, OT_NULL, integer, 0, simplify)
#define XREF__FUNCTION(name, type, field, ot, pst, major, minor, mand, ser) \
XREF__FUNCTION_DATA(name, type, field, ot, pst, NULL, \
                   major, minor, mand, ser, 0)

#define XREF_S_ARRAY_DATA(name, type, field, pst, data, major, minor, mand) \
XREF__ARRAY (name, type, field, OT_ANY, pst, data, \
             major, minor, mand, XREF_S_INLINE, 1)
#define XREF_A_ARRAY_DATA(name, type, field, pst, data, major, minor, mand) \
XREF__ARRAY (name, type, field, OT_ARRAY, pst, data, \
             major, minor, mand, XREF_S_INLINE, 0)

#define XREF_S_ARRAY(name, type, field, pst, major, minor, mand) \
XREF_S_ARRAY_DATA (name, type, field, pst, NULL, major, minor, mand)

#define XREF_A_ARRAY(name, type, field, pst, major, minor, mand) \
XREF_A_ARRAY_DATA (name, type, field, pst, NULL, major, minor, mand)


#define XREF_FUNCTION_R(name, type, field, pst, major, minor, mand) \
XREF__FUNCTION (name, type *, field, OT_ANY, pst, major, minor, \
                mand, XREF_S_REFERENCE)
#define XREF_FUNCTION_R_DATA(name, type, field, pst, data, \
                             major, minor, mand) \
XREF__FUNCTION_DATA (name, type *, field, OT_ANY, pst, data, major, minor, \
                     mand, XREF_S_REFERENCE)

#define XREF_FUNCTION_I(name, type, field, pst, major, minor, mand) \
XREF__FUNCTION (name, type, field, OT_ANY, pst, major, minor, \
                mand, XREF_S_INLINE)
#define XREF_FUNCTION_I_DATA(name, type, field, pst, data, \
                             major, minor, mand) \
XREF__FUNCTION_DATA (name, type, field, OT_ANY, pst, data, major, minor, \
                     mand, XREF_S_INLINE, 0)

#define XREF_INCLUDE(type, field, ot, fd) \
XREF__F ("", type, field, ot, XREF_D_NEST, XREF_S_INLINE, \
         1, 0, XREF_M, fdesc, fd, NULL, OT_NULL, integer, 0, 0)
#define XREF_NEST(name, type, field, ot, fd, major, minor, mand) \
XREF__F (name, type, field, ot, XREF_D_NEST, XREF_S_INLINE, \
         major, minor, mand, fdesc, fd, NULL, OT_NULL, integer, 0, 0)
#define XREF_REGION(field) \
XREF__F ("", struct xref_stream, field, OT_STREAM, XREF_D_REGION, \
         XREF_S_INLINE, 1, 0, XREF_M, pstable, NULL, NULL, \
         OT_NULL, integer, 0, 0)
#define XREF_DYN(type, field, dyf) \
XREF__F ("", type, field, OT_ANY, XREF_D_DYNAMIC, XREF_S_INLINE, \
         1, 0, XREF_M, dyn, dyf, NULL, OT_NULL, integer, 0, 0)

#ifdef XREF_IMPLEMENTATION
#define XREF_DESC_BEGIN(name, val)  \
static struct xref_field_desc name = { val, {
#define XREF_DESC_END { NULL }  } };
#else
#define XREF_DESC_BEGIN(name, val)
#define XREF_DESC_END
#endif

#define XREF_F_DICTIONARY(name, fd, major, minor) \
XREF__D (name, OT_DICTIONARY, XREF_S_INLINE, XREF_D_NEST, \
	 major, minor, fdesc, &fd, NULL, 0)

#define XREF_F_STREAM(name, fd, major, minor) \
XREF__D (name, OT_STREAM, XREF_S_INLINE, XREF_D_NEST, \
         major, minor, fdesc, &fd, NULL, 0)

#define XREF_F_S_ARRAY(name, fd, major, minor) \
XREF__D (name, OT_ARRAY, XREF_S_INLINE, XREF_D_NEST, \
	 major, minor, fdesc, &fd, NULL, 1)

#define XREF_F_A_ARRAY(name, fd, major, minor) \
XREF__D (name, OT_ARRAY, XREF_S_INLINE, XREF_D_NEST, \
	 major, minor, fdesc, &fd, NULL, 0)

#define XREF_F_INTEGER(name, major, minor) \
XREF__D (name, OT_INTEGER, XREF_S_INLINE, XREF_D_UNWRAP, \
	 major, minor, pstable, NULL, NULL, 0)

#define XREF_F_FUNCTION(name, pst, major, minor) \
XREF__D (name, OT_ANY, XREF_S_INLINE, XREF_D_PSTABLE, major, minor, \
	 pstable, &pst, NULL, 0)

#define XREF_F_DYN(name, dyf, major, minor) \
XREF__D (name, OT_ANY, XREF_S_INLINE, XREF_D_DYNAMIC, major, minor, \
	 dyn, dyf, NULL, 0)

