/*
 * PDFlow -- machine.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef machine_h

#define machine_h

#include "status.h"
#include "fmt/obj.h"

#define M_TOK_BLOCK -1

struct m_rule;
struct m_machine;
struct m_stack;
struct m_state;

typedef enum s_status (*m_trigger) (struct m_stack *stack,
				    struct obj *obj, int token,	
				    void *closure);


struct m_block_desc {

	int open;

	int close;
};

struct m_rule {

	int token;

	m_trigger trigger;
};

enum s_status m_create_machine (struct m_rule *rules, int toks,
				struct m_block_desc *block, int blocks,
				int eoftok, struct m_machine **machp);
void m_destroy_machine (struct m_machine *mach);

enum s_status m_start_machine (struct m_machine *mach,
			       struct m_state **state);
enum s_status m_open_machine (struct m_state *state);
void m_close_machine (struct m_state *state);
void m_stop_machine (struct m_state *state);

int m_run_machine_cyc (struct m_state *state, struct obj *obj, int token,
		       enum s_status *res, void *closure);
enum s_status m_eval_block (struct m_state *state, struct obj *obj,
			    void *closure);
int m_compare_blocks (struct obj *obj1, struct obj *obj2);
enum s_status m_push_stack (struct m_state *state, struct obj *obj, int tt,
			    int count);
enum s_status m_dump_stack (struct m_state *state, struct obj *obj, 
			    int *count);

int m_empty (struct m_stack *stack);
unsigned m_size (struct m_stack *stack);
enum s_status m_push (struct m_stack *stack, struct obj *obj, int tt);
int m_pop (struct m_stack *stack, struct obj *obj, int *tt);
enum s_status m_copy (struct m_stack *stack, struct obj *refobj, 
		      pdf_integer_t n);
enum s_status m_roll (struct m_stack *stack, struct obj *refobj, 
		      pdf_integer_t j, pdf_integer_t n);
enum s_status m_index (struct m_stack *stack, struct obj *refobj, 
		       pdf_integer_t n);
enum s_status m_exch (struct m_stack *stack, struct obj *refobj);
struct obj *m_top (struct m_stack *stack, int *tt);
void m_open (struct m_stack *stack);
void m_close (struct m_stack *stack);


#endif
