/*
 * PDFlow -- pdf-lex.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef __pdf_lex_h

#define __pdf_lex_h

#include <stdarg.h>
#include "fmt/obj.h"
#include "fmt/file.h"

enum pl_token_type {
	TOK_EOF,

	TOK_PLUS,
	TOK_MINUS,
	TOK_ARRAY_BEGIN,
	TOK_ARRAY_END,
	TOK_NULL,
	TOK_TRUE,
	TOK_FALSE,
	TOK_BOOLEAN,
	TOK_INTEGER,
	TOK_REAL,
	TOK_NAME,
	TOK_LITERAL,
	TOK_HEX,
	TOK_STREAM,
	TOK_ENDSTREAM,
	TOK_OBJ,
	TOK_ENDOBJ,
	TOK_DICT_BEGIN,
	TOK_DICT_END,
	TOK_REF,
	TOK_N,
	TOK_F,
	TOK_TRAILER,
	TOK_STARTXREF,

	TOK_KEYWORD,
	TOK_OTH_KEYWORD,
	TOK_LCURLY,
	TOK_RCURLY,

	/* Composite objects, never returned by the lexer */
	TOK_ARRAY,
	TOK_DICTIONARY,
	TOK_REFERENCE,
	TOK_DEF_REFERENCE,
	TOK_DEFINITION,
	TOK_XREF,

	TOK__LAST__
};

struct pl_state {

	struct f_ptr ptr, sptr;

	struct obj_keyword *keys;

	int eof;

	struct obj token;
};

int pl_is_blank (char c);
int pl_skip_ws (struct pl_state *ps);
void pl_init (struct pl_state *ps, struct f_ptr *ptr);
void pl_init_kw (struct pl_state *ps, struct f_ptr *ptr, 
		 struct obj_keyword *keys);
enum s_status pl_seek_bol (struct pl_state *ps, const char *key);
enum s_status pl_lex (struct pl_state *ps, enum pl_token_type *tt);
enum s_status pl_lex_predict (struct pl_state *ps, enum pl_token_type tt,
			      const char *tag);
enum s_status pl_backline (struct pl_state *ps, unsigned n);
enum s_status pl_verror (struct pl_state *ps, const char *fmt, va_list ap);
enum s_status pl_error (struct pl_state *ps, const char *fmt, ...);

enum s_status pl_serialize (struct f_file *file, struct obj *obj);

enum s_status pl_get_header (struct f_file *file, const char *name,
			     char **result);

#endif
