/*
 * PDFlow -- xref.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef xref_h

#define xref_h

/* Do not include this file directly, but use fields.h instead. 
 * This beacuse of the strange circula dependency behind the
 * definition of a stream */

#include "version.h"
#include "status.h"
#include "obj.h"
#include "refs.h"

struct xref_db;
struct xref_bind;
struct f_file;

typedef void (*xref_cache_unref) (void *ptr, struct ref_owner *p, 
				  void *closure);
struct xref_stream {
	
	struct obj_region encoded;
	struct obj_region decoded;

#define XREF_STREAM
#include "xref.tab.h"	
};

enum s_status xref_create_db (struct xref_db **prdb, struct ref_owner *p);
enum s_status xref_init_db (struct xref_db *rdb, struct f_file *f);
enum s_status xref_scratch_db (struct xref_db *rdb, struct v_selection *sel,
			       struct f_file *file);
enum s_status xref_sync_start (struct xref_db *rdb, struct v_selection *sel);
enum s_status xref_sync_db (struct xref_db *rdb, struct v_selection *sel,
			    struct obj *trailer);
enum s_status xref_get_file_info (struct xref_db *rdb, 
				  const char **fname, pdf_ofs_t *size);
enum s_status xref_get_object (struct xref_db *db, 
			       struct obj_reference *ref,
			       struct obj *obj);
enum s_status xref_register_object (struct xref_db *rdb, struct obj *obj);
enum s_status xref_remove_object (struct xref_db *rdb, struct obj *obj);

enum s_status xref_resolve_no_cast (struct xref_db *rdb, struct obj *sobj,
				    enum obj_type ot, struct obj *dobj, 
				    int *present, const char *label);
enum s_status xref_resolve (struct xref_db *rdb, struct obj *sobj,
			    enum obj_type ot, struct obj *dobj, 
			    int *present, const char *label);
enum s_status xref_resolve_stream (struct xref_db *rdb,
				   struct obj *sobj);
enum s_status xref_resolve_stream_xs (struct xref_db *rdb,
				      struct obj *sobj,
				      struct xref_stream *xs,
				      struct ref_owner *p);
void xref_stream_destroy (struct xref_stream *xs, struct ref_owner *p);
enum s_status xref_stream_ser (struct xref_db *rdb, struct xref_bind *robj,
			       struct v_selection *sel, struct obj *obj,
			       struct xref_stream *xs);
enum s_status xref_serialize_obj (struct xref_db *rdb, struct obj *obj);

void xref_set_aborted (struct xref_db *rdb);

void xref_ref_db (struct xref_db *rdb, struct ref_owner *p);
void xref_unref_db (struct xref_db *rdb, struct ref_owner *p);

enum s_status xref_add_selection (struct v_selection *sel);

struct obj *xref_get_trailer (struct xref_db *rdb);

pdf_integer_t xref_get_id (struct xref_db *rdb);

enum s_status xref_cache_add (struct xref_db *rdb, const char *id,
			      xref_cache_unref unref, void *closure);
void xref_cache_remove (struct xref_db *rdb, const char *id);
struct xref_cache *xref_cache_get (struct xref_db *rdb, const char *id);
void *xref_cache_find (struct xref_cache *cache, const struct obj *obj);
struct ref_owner *xref_cache_put (struct xref_cache *cache, 
				  const struct obj *obj, void *ptr);


#endif
