/*
 * PDFlow -- 2d.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef twod_h

#define twod_h

#include "graph/agc.h"

enum td_space_type {

	TD_ST_GRAPHICAL,

	TD_ST_ORDINARY

};

struct td_space {

	enum td_space_type type;

	pdf_real_t amount;

};

struct td_graph_spaces {
	pdf_real_t left, right, top, bottom;
};

struct td_line_status {

	struct td_float *tdf;

	struct gs_point point;

	struct gs_point wvec, lvec;

	pdf_bool_t wpi, lpi;

};

struct td_float {

	ref_t refs;

	struct td_float *next;

	struct td_graph_spaces spaces;

	enum gs_direction wdir, ldir;

	struct cobj_rect rect, avail;

	struct gs_point origin;

	struct agc_context *agc;
};

struct td_pool {

	ref_t refs;

	struct cobj_rect total, avail;

	struct td_float *floats;

};

enum s_status td_pool__new (struct td_pool **tdp, struct cobj_rect *rect,
			    struct ref_owner *p, const char *fname, int line);
#define td_pool_new(tdp, rect, p) \
td_pool__new (tdp, rect, p, __FILE__, __LINE__)

void td_pool__ref (struct td_pool *tdp, struct ref_owner *p,
		   const char *fname, int line);
#define td_pool_ref(tdp, p) td_pool__ref (tdp, p, __FILE__, __LINE__)
void td_pool_unref (struct td_pool *tdp, struct ref_owner *p);

void td_pool_avail (struct td_pool *tdp, struct cobj_rect *rect);
void td_pool_line_avail (struct td_pool *tdp, struct cobj_rect *rect);
enum s_status td_pool_finalize (struct td_pool *tdp, struct agc_context **agcp,
				struct ref_owner *p);

enum s_status td_float_alloc (struct td_float **tdf, struct td_pool *tdp,
			      enum gs_direction wdir, enum gs_direction ldir,
			      struct td_space *space);
void td_float_done (struct td_pool *tdp, struct td_float *tdf, 
		    struct td_space *space);

enum s_status td_float_bol (struct td_line_status *tdl, struct td_float *tdf, 
			    pdf_real_t ascent, pdf_real_t descent);
enum s_status td_float_write (struct td_line_status *tdl, 
			      struct agc_context *agc,
			      struct cobj_rect *rect, struct gs_point *origin);
enum s_status td_float_space (struct td_line_status *tdl, pdf_real_t amount);
enum s_status td_float_eol (struct td_line_status *tdl);

#endif
