/*
 * PDFlow -- agc.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef agc_h

#define agc_h

#include "graph/gspg.h"

struct agc_text {

	struct gs_paint paint;
	
	struct obj_array text;

	struct obj_array_node *tail;

	struct cobj_matrix tm0, tm;
};

struct agc_page_context {

	struct gs_context *gpc;

	struct gspg_page *gspg;

	struct agc_text at;

};

typedef int agc_code_t;

struct agc_float_command {

	struct agc_float_command *next;

	enum agc_float_command_type {

		AGC_FTC_RGB,
		AGC_FTC_RECT,
		AGC_FTC_SET_STATE,
		AGC_FTC_WRITE_CHAR,
		AGC_FTC_CTM,
		AGC_FTC_MOVE_TO,
		AGC_FTC_LINE_TO,
		AGC_FTC_BEZIER_TO,
		AGC_FTC_STROKE,
		AGC_FTC_XOBJ,
		AGC_FTC_SHADING

	} type;
	
	pdf_bool_t enabled;

	agc_code_t code;

	union {

		struct agc_ftc_rgb {
			pdf_real_t r, g, b;
			pdf_bool_t stroking;
		} rgb;

		struct agc_ftc_rect {
			struct cobj_rect rect;
			struct gs_paint paint;
		} rect;

		struct agc_ftc_set_state {
			struct gs_text_state *tstate;
			struct gs_graph_state *gstate;
			struct cobj_rect cpath;
		} set_state;

		struct agc_ftc_write_char {
			struct gs_glyph_info gi;
			struct gs_paint paint;
		} write_char;

		struct agc_ctm {
			struct cobj_matrix mat;
		} ctm;

		struct agc_move_to {
			struct gs_point p;
		} move_to;

		struct agc_line_to {
			struct gs_point p;
		} line_to;

		struct agc_bezier_to {
			struct gs_point b, ca, cb;
		} bezier_to;

		struct agc_stroke {
			struct gs_paint paint;
		} stroke;

		struct agc_xo {
			struct xo_object *xo;
		} xobj;

		struct agc_shading {
			struct shd_shading *sh;
		} shading;

	} d;
};

struct agc_float_context {

	struct agc_float_command *cmd;

};

struct agc_measure_context {

	struct cobj_matrix ctm;

	struct cobj_rect rect;

	struct cobj_rect cpath;

	struct cobj_rect path;

	struct gs_point point;

};

struct agc_context;

struct agc_ifc {

	enum s_status (*done) (struct agc_context *agc);

	void (*destroy) (struct agc_context *agc);

	enum s_status (*rgb) (struct agc_context *agc, pdf_real_t r, 
			      pdf_real_t g, pdf_real_t b, pdf_bool_t stroking);

	enum s_status (*ctm) (struct agc_context *agc, 
			      struct cobj_matrix *ctm);

	enum s_status (*rect) (struct agc_context *agc, 
			       struct cobj_rect *rect, struct gs_paint *paint);

	enum s_status (*move_to) (struct agc_context *agc, 
				  struct gs_point *p);

	enum s_status (*line_to) (struct agc_context *agc, 
				  struct gs_point *p);

	enum s_status (*bezier_to) (struct agc_context *agc, 
				    struct gs_point *b,
				    struct gs_point *ca,
				    struct gs_point *cb);

	enum s_status (*stroke) (struct agc_context *agc, 
				 struct gs_paint *paint);

	enum s_status (*xo) (struct agc_context *agc, 
			     struct xo_object *xo);

	enum s_status (*shading) (struct agc_context *agc, 
				  struct shd_shading *sh);
	
	enum s_status (*set_state) (struct agc_context *agc,
				    struct gs_text_state *tstate,
				    struct gs_graph_state *gstate,
				    struct cobj_rect *cpath);

	enum s_status (*write_char) (struct agc_context *agc,
				     struct gs_glyph_info *gi,
				     struct gs_paint *paint);

	enum s_status (*play) (struct agc_context *agc,
			       struct agc_context *pagc);

	void (*set_enable) (struct agc_context *agc,
			    agc_code_t code, pdf_bool_t enabled);
};

struct agc_context {

	ref_t refs;

	struct agc_ifc *ifc;

	struct cobj_matrix bctm;

	agc_code_t current_code;

	union {

		struct agc_page_context pg;

		struct agc_float_context ft;

		struct agc_measure_context mt;

	} d;
};

enum s_status agc__new (struct agc_context **agc, struct xref_db *rdb,
			struct gspg_page *gspg, struct ref_owner *p, 
			const char *fname, int line);
#define agc_new(agc, rdb, gspg, p) \
agc__new (agc, rdb, gspg, p, __FILE__, __LINE__)

enum s_status agc__float_new (struct agc_context **agc, struct ref_owner *p,
			      const char *fname, int line);
#define agc_float_new(agc, p) \
agc__float_new (agc, p, __FILE__, __LINE__)

enum s_status agc__measure_new (struct agc_context **agc, struct ref_owner *p,
				const char *fname, int line);
#define agc_measure_new(agc, p) \
agc__measure_new (agc, p, __FILE__, __LINE__)

void agc__ref (struct agc_context *agc, struct ref_owner *p, 
	       const char *fname, int line);
#define agc_ref(agc, p) agc__ref (agc, p, __FILE__, __LINE__)

void agc_unref (struct agc_context *agc, struct ref_owner *p);

enum s_status agc_done (struct agc_context *agc);

enum s_status agc_rgb (struct agc_context *agc, pdf_real_t r, 
		       pdf_real_t g, pdf_real_t b);
enum s_status agc_rgb_s (struct agc_context *agc, pdf_real_t r, 
			 pdf_real_t g, pdf_real_t b, pdf_bool_t stroking);
enum s_status agc_ctm (struct agc_context *agc, struct cobj_matrix *mat);
enum s_status agc_rect (struct agc_context *agc, 
			struct cobj_rect *rect, struct gs_paint *paint);
enum s_status agc_triangle (struct agc_context *agc,
			    struct gs_point *a,
			    struct gs_point *b,
			    struct gs_point *c,
			    struct gs_paint *paint);

enum s_status agc_draw_subpath (struct agc_context *agc,
				struct gs_subpath *sp,
				struct gs_paint *paint);
enum s_status agc_move_to (struct agc_context *agc,
			   struct gs_point *p);
enum s_status agc_line_to (struct agc_context *agc,
			   struct gs_point *p);
enum s_status agc_bezier_to (struct agc_context *agc,
			     struct gs_point *b,
			     struct gs_point *ca,
			     struct gs_point *cb);
enum s_status agc_stroke (struct agc_context *agc,
			  struct gs_paint *paint);

enum s_status agc_draw_xobject (struct agc_context *agc,
				struct xo_object *xo);

enum s_status agc_draw_shading (struct agc_context *agc,
				struct shd_shading *sh);

enum s_status agc_set_state (struct agc_context *agc,
			     struct gs_text_state *tstate,
			     struct gs_graph_state *gstate,
			     struct cobj_rect *cpath);

enum s_status agc_write_char (struct agc_context *agc,
			      struct gs_glyph_info *gi,
			      struct gs_paint *paint);

enum s_status agc_translate_like (struct agc_context *agc,
				  struct cobj_rect *frect,
				  struct cobj_rect *trect);

enum s_status agc_apply (struct agc_context *agc,
			 struct agc_context *pagc,
			 struct gs_point *point);

enum s_status agc_measure (struct agc_context *agc, struct cobj_rect *rect);

void agc_set_code (struct agc_context *agc,
		   agc_code_t code);

void agc_unset_code (struct agc_context *agc);

void agc_enable_code (struct agc_context *agc,
		      agc_code_t code);

void agc_disable_code (struct agc_context *agc,
		       agc_code_t code);

#endif
