/*
 * PDFlow -- gspg.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef gspg_h

#define gspg_h

#include "fmt/fields.h"
#include "types/page.h"
#include "types/xobj.h"
#include "types/catalog.h"
#include "graph/gstate.h"

enum gspg_stmt_type {

	GSPG_ST_PATH,
	GSPG_ST_SHADING,
	GSPG_ST_XOBJECT,
	GSPG_ST_INLINE_IMAGE

};

struct gspg_stmt {

	struct gspg_stmt *next;

	enum gspg_stmt_type type;

	struct gs_graph_state *gstate;

	struct gs_text_state *tstate;

	pdf_bool_t no_cpath;

	union {
		struct gs_path *path;
		struct shd_shading *shading;
		struct xo_object *xobject;
	} u;

	struct gs_paint paint;
};

struct gspg_page {

	ref_t refs;

	struct cobj_rect bounds;

	pdf_integer_t rotate;

	int number;

	struct gspg_stmt *head, *tail;

};

enum s_status gspg_stmt_create (struct gspg_stmt **stmt);
void gspg_stmt_destroy (struct gspg_stmt *stmt, struct ref_owner *p);

enum s_status gspg_stmt_add (struct gspg_page *page,
			     struct gs_context *gc,
			     struct gspg_stmt **stmtp);

enum s_status gspg_stmt_add_path (struct gspg_page *page, 
				  struct gs_context *gc,
				  struct gs_paint *paint,
				  pdf_bool_t no_cpath);

enum s_status gspg_stmt_add_xobj (struct gspg_page *page, 
				  struct gs_context *gc,
				  struct xo_object *xo,
				  pdf_bool_t no_cpath);

enum s_status gspg_stmt_add_shading (struct gspg_page *page, 
				     struct gs_context *gc,
				     struct shd_shading *sh,
				     pdf_bool_t no_cpath);

enum s_status gspg_page_new (struct gspg_page **page, struct cobj_rect *bounds,
			     pdf_integer_t rotate, struct ref_owner *p);
void gspg_page_ref (struct gspg_page *page, struct ref_owner *p);
void gspg_page_unref (struct gspg_page *page, struct ref_owner *p);

enum s_status gspg_expand_shorthands (struct obj_dictionary *dict);
enum s_status gspg_apply_shorthands (struct obj_dictionary *dict);

#endif
