/*
 * PDFlow -- gstate.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef gstate_h

#define gstate_h

#include "refs.h"
#include "fmt/fields.h"
#include "types/pattern.h"
#include "types/colour.h"
#include "fmt/obj.h"
#include "types/font.h"
#include "types/tobj.h"

extern struct cobj_enum_desc gs_rendering_intent_desc [];

typedef unsigned long gs_keyset_t;

enum gs_render_type {

	GS_RT_FILL = 0,
	GS_RT_STROKE = 1,
	GS_RT_FILL_STROKE = 2,
	GS_RT_NONE = 3,
	GS_RT_FILL_CLIP = 4,
	GS_RT_STROKE_CLIP = 5,
	GS_RT_FILL_STROKE_CLIP = 6,
	GS_RT_CLIP = 7

};

enum gs_black_generation_type {

	GS_BG_EXPLICIT_DEFAULT,

	GS_BG_FUNCTION,
	GS_BG_DEFAULT
};

enum gs_undercolor_removal_type {

	GS_UCR_EXPLICIT_DEFAULT,

	GS_UCR_FUNCTION,
	GS_UCR_DEFAULT
};

enum gs_transfer_function_type {
	GS_TR_IDENTITY,
	GS_TR_EXPLICIT_DEFAULT,

	GS_TR_FUNCTION,
	GS_TR_DEFAULT
};

enum gs_halftone_dictionary_type {
	GS_HT_EXPLICIT_DEFAULT,

	GS_HT_FUNCTION,
	GS_HT_DEFAULT
};

enum gs_blend_mode_type {
	GS_BM_NORMAL,
	GS_BM_COMPATIBLE,
	GS_BM_MULTIPLY,
	GS_BM_SCREEN,
	GS_BM_OVERLAY,
	GS_BM_DARKEN,
	GS_BM_LIGHTEN,
	GS_BM_COLOR_DODGE,
	GS_BM_COLOR_BURN,
	GS_BM_HARD_LIGHT,
	GS_BM_SOFT_LIGHT,
	GS_BM_DIFFERENCE,
	GS_BM_EXCLUSION,
	
	GS_BM_HUE,
	GS_BM_SATURATION,
	GS_BM_COLOR,
	GS_BM_LUMINOSITY
};

enum gs_soft_mask_type {

	GS_SM_NONE,

	GS_SM_DICTIONARY,
	GS_SM_DEFAULT
};

enum gs_element_type {

	GS_ET_LINE,
	GS_ET_BEZIER,
	GS_ET_TEXT

};

enum gs_fill_rule {

	GS_FR_NONE,

	GS_FR_WINDING,
	GS_FR_EVEN_ODD

};

enum gs_line_cap {
	GS_LC_BUTT = 0,
	GS_LC_ROUND = 1,
	GS_LC_SQUARE = 2
};

enum gs_line_join {
	GS_LJ_MITER = 0,
	GS_LJ_ROUND = 1,
	GS_LJ_BEVEL = 2
};

enum gs_rendering_intent_type {
	GS_RI_ABSOLUTE_COLOMETRIC,
	GS_RI_RELATIVE_COLOMETRIC,
	GS_RI_SATURATION,
	GS_RI_PERCEPTUAL,
	
	GS_RI_OTHER
};

enum gs_state_op {
	GS_GO_OP,
	GS_GO_COLOUR,
	GS_GO_DICT,
	GS_GO_CLIPPING_PATH
};

enum gs_direction {
	GS_D_UNKNOWN,

	GS_D_LEFT_TO_RIGHT,
	GS_D_RIGHT_TO_LEFT,
	GS_D_TOP_TO_BOTTOM,
	GS_D_BOTTOM_TO_TOP,

	GS_D__LAST__
};

T_PROTOTYPES (gs_graph, gs_graph_state);
T_PROTOTYPES (gs_text, gs_text_state);
T_PROTOTYPES (gs, gs_state);
T_PROTOTYPES (gs_dash_pattern, gs_dash_pattern);
T_PROTOTYPES (gs_font, gs_font);
T_PROTOTYPES (gs_context, gs_context);

#define GS_DASH_PATTERN_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct gs_dash_pattern, field, \
                 &gs_dash_pattern_pstable, major, minor, mand)

#define GS_FONT_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct gs_font, field, \
                 &gs_font_pstable, major, minor, mand)

#define GS_GRAPH_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct gs_graph_state, field, \
                 &gs_graph_pstable, major, minor, mand)

#define GS_TEXT_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct gs_text_state, field, \
                 &gs_text_pstable, major, minor, mand)

#define GS_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct gs_state, field, \
                 &gs_pstable, major, minor, mand)

typedef pdf_real_t (*gs_edge) (struct cobj_rect *rect);
typedef pdf_real_t (*gs_cmp) (pdf_real_t a, pdf_real_t b);

struct gs_paint {

	pdf_bool_t stroke;

	pdf_bool_t fill;

	enum gs_fill_rule rule;
};

extern struct gs_paint gs_paint_stroke;
extern struct gs_paint gs_paint_fill;
extern struct gs_paint gs_paint_stroke_fill;

struct gs_point {

	pdf_real_t x, y;
};

struct gs_mat {
	struct cobj_matrix t_m;
	struct cobj_matrix t_lm;
};

struct gs_font {

	struct t_type t;

#define GS_FONT
#include "graph/gstate.tab.h"	
};

struct gs_state_key {

	struct obj_bstr opkey;

	struct obj_bstr dkey;	
};

struct gs_text_state {

	struct t_type t;

	int frozen;

	struct gs_state_key key;

	pdf_real_t char_space;

	pdf_real_t word_space;

	pdf_real_t scale;

	pdf_real_t leading;

	enum gs_render_type render;

	pdf_real_t rise;

	struct gs_mat mat;

#define GS_TEXT_STATE
#include "graph/gstate.tab.h"	
};

struct gs_dash_pattern {

	struct t_type t;

#define GS_DASH_PATTERN
#include "graph/gstate.tab.h"	
};

struct gs_clipping_path {

	ref_t refs;

	struct gs_clipping_path *next;

	struct gs_path *path;

	enum gs_fill_rule rule;

	struct gs_graph_state *gstate;

	struct gs_text_state *tstate;
};

struct gs_graph_state {
	
	struct t_type t;

	int frozen;

	struct gs_state_key key;

	struct cobj_matrix ctm;

	struct gs_clipping_path *clipping_path;

	struct gc_colour_state *stroking_colour_state;

	struct gc_colour_state *nonstroking_colour_state;;

#define GS_GRAPH_STATE
#include "gstate.tab.h"
};

struct gs_state {

	struct t_type t;

	struct gs_state *next;

#define GS_GSTATE
#include "gstate.tab.h"
};

struct gs_line {
	struct gs_point a, b;
};

struct gs_bezier {
	struct gs_point a, b;
	struct gs_point ca, cb;
};

struct gs_sh {
	struct pt_pattern pattern;
};

struct gs_text {
	struct obj text;

	struct gs_text_state *tstate;

	struct gs_graph_state *gstate;
};

struct gs_char_info {
	size_t size;
	char bytes [16];
};

struct gs_glyph_info {
	struct gf_glyph_info f;
	struct gs_char_info c;
	struct cobj_matrix t_m;
};

struct gs_text_string {
	int elements;
	struct gs_glyph_info *glyphs;
};

struct gs_element {

	enum gs_element_type type;
	
	union {
		struct gs_line line;
		struct gs_bezier bezier;
		struct gs_text text;
	} u;
};

struct gs_subpath {
	struct gs_subpath *next;

	int elements, cap;

	struct gs_element *e;
};

struct gs_path {

	ref_t refs;

	struct gs_subpath *head;
};

struct gs_context {

	struct t_type t;

	struct gs_state *state;
	
	struct gs_path *path;

	struct gs_subpath *gss;

	int sb_path_closed;

	struct gs_point start_point, curr_point;
};

struct gs_tags {
	struct t_tag halftone;
	struct t_tag soft_mask;
};

enum s_status gs_subpath_copy (struct gs_subpath *dsp, struct gs_subpath *ssp, 
			       struct ref_owner *p);
void gs_subpath_destroy (struct gs_subpath *sp, struct ref_owner *p);
int gs_point_same (struct gs_point *a, struct gs_point *b);

enum s_status gs_set_ctm (struct gs_context *gpc, struct cobj_matrix *coef);

enum s_status gs_add_clipping_path (struct gs_context *gpc,
				    struct gs_path *path,
				    enum gs_fill_rule rule);
void gs_unref_clipping_path (struct gs_graph_state *gss);

enum s_status gs_dash_pattern_init_alt (struct gs_dash_pattern **dp,
					struct xref_db *rdb,
					struct xref_bind *robj,
					pdf_integer_t phase,
					const struct obj *array,
					struct ref_owner *p);
enum s_status gs_context_set_dash_pattern_alt (struct gs_context *gc,
					       pdf_integer_t phase,
					       const struct obj *array);
int gs_dash_pattern_eq (struct gs_dash_pattern *dp1, 
			struct gs_dash_pattern *dp2);
int gs_rendering_intent_eq (struct fun_enum *ri1,
			    struct fun_enum *ri2);
enum s_status gs_context_set_rendering_intent (struct gs_context *gc,
					       struct obj *obj);
enum s_status gs_context_set_dictionary (struct gs_context *gc,
					 struct obj *dict,
					 struct gs_tags *tags);
enum s_status gs_context_set_font (struct gs_context *gpc, 
				   struct gf_font *font, pdf_real_t size);

struct ref_owner *gs_text_owner (struct gs_context *gc);
struct ref_owner *gs_graph_owner (struct gs_context *gc);

enum s_status gs_state_key_dup (struct gs_state_key *skey, 
				struct gs_state_key *dkey);
void gs_state_key_destroy (struct gs_state_key *key);
int gs_state_key_compare (const struct gs_state_key *key1,
			  const struct gs_state_key *key2);
pdf_bool_t gs_graph_equal (const struct gs_graph_state *gss1,
			   const struct gs_graph_state *gss2);
pdf_bool_t gs_text_equal (const struct gs_text_state *tws1,
			  const struct gs_text_state *tws2);

enum s_status gs_state_dkey_dup (struct gs_state_key *skey, 
				 struct gs_state_key *dkey);
int gs_state_dkey_compare (struct gs_state_key *key1,
			   struct gs_state_key *key2);

enum s_status gs_filter_gs (struct obj *obj, gs_keyset_t *keyset);

#ifdef DEBUG
void gs_text_check_key (struct gs_text_state *tws);
#else
#define gs_text_check_key(x)
#endif
enum s_status gs_text_update_key (struct gs_text_state *tws, 	
				  enum gs_state_op op);
enum s_status gs_text (struct gs_context *gc, enum gs_state_op op,
		       struct gs_text_state **tws);
const struct gs_text_state *gs_ro_text (struct gs_context *gc);
struct gs_text_state *gs_snap_text (struct gs_context *gc,
				    struct ref_owner *p);
enum s_status gs_set_tstate (struct gs_context *gc,
			     struct gs_text_state *tws);
enum s_status gs_import_tstate (struct gs_text_state **tws,
				struct ref_owner *p);
pdf_bool_t gs_tstate_eq (const struct gs_text_state *twsa,
			 const struct gs_text_state *twsb);
#ifdef DEBUG
void gs_graph_check_key (struct gs_graph_state *gss);
#else
#define gs_graph_check_key(x)
#endif
enum s_status gs_graph_update_key (struct gs_graph_state *gss, 
				   enum gs_state_op op);
enum s_status gs_graph (struct gs_context *gc, enum gs_state_op op,
			struct gs_graph_state **gss);
const struct gs_graph_state *gs_ro_graph (struct gs_context *gc);
struct gs_graph_state *gs_snap_graph (struct gs_context *gc,
				      struct ref_owner *p);
enum s_status gs_set_gstate (struct gs_context *gc,
			     struct gs_graph_state *gss);
enum s_status gs_import_gstate (struct gs_graph_state **gss,
				struct ref_owner *p);

enum s_status gs_push (struct gs_context *gc);
enum s_status gs_pop (struct gs_context *gc, int *eos);
enum s_status gs_done (struct gs_context *gc);

enum s_status gs_path_reset (struct gs_context *gpc);
void gs_path_ref (struct gs_path *gp, struct ref_owner *p);
void gs_path_unref (struct gs_path *gp, struct ref_owner *p);

enum s_status gs_move (struct gs_context *gpc, struct obj *ref,
			pdf_real_t x, pdf_real_t y);
enum s_status gs_lineto (struct gs_context *gpc, struct obj *ref,
			  pdf_real_t x, pdf_real_t y);
enum s_status gs_bezierto (struct gs_context *gpc, struct obj *ref,
			    pdf_real_t cx1, pdf_real_t cy1,
			    pdf_real_t cx2, pdf_real_t cy2,
			    pdf_real_t x, pdf_real_t y);
enum s_status gs_close (struct gs_context *gpc, struct obj *ref);
enum s_status gs_rect (struct gs_context *gpc, struct obj *ref,
		       struct cobj_rect *rect);

enum s_status gs_begin_text (struct gs_context *gpc, const struct obj *ref);
enum s_status gs_newline (struct gs_context *gpc, 
			   pdf_real_t tx, pdf_real_t ty);
enum s_status gs_set_mat (struct gs_context *gpc, struct cobj_matrix *mat);
enum s_status gs_write_text (struct gs_context *gpc, struct gs_text *text);
enum s_status gs_write (struct gs_context *gpc, struct obj *text);

pdf_bool_t gs_get_kerning (struct gs_context *gpc, 
			   struct cobj_matrix *otm, struct cobj_matrix *ntm,
			   pdf_real_t *kerning);

int gs_t_clip (struct gs_context *gpc);

void gs_paint_to_render (struct gs_paint *paint, 
			 enum gs_render_type *rt, enum gs_fill_rule cfrule);
void gs_render_to_paint (enum gs_render_type rt, struct gs_paint *paint);

enum s_status gs_set_device_colour (struct gs_context *gc,
				    struct rd_resdict *rd,
				    struct gc_colour_state **cst,
				    enum gc_colour_space_family family);
enum s_status gs_set_stroking_colour (struct gs_context *gc,
				      enum gc_colour_space_family family,
				      pdf_real_t *data);
enum s_status gs_set_nonstroking_colour (struct gs_context *gc,
					 enum gc_colour_space_family family,
					 pdf_real_t *data);

enum s_status gs_tags_get (struct gs_graph_state *gs, struct gs_tags *tags);
enum s_status gs_tags_dup (struct gs_tags *stags, struct gs_tags *dtags);
void gs_tags_destroy (struct gs_tags *tags);

void gs_translate_vector (struct gs_point *point, struct cobj_matrix *mat);
void gs_translate_coords (struct gs_point *point, struct cobj_matrix *mat);
void gs_translate_rect (struct cobj_rect *rect, struct cobj_matrix *mat);
void gs_revert_coords (struct gs_point *point, struct cobj_matrix *mat);
void gs_revert_rect (struct cobj_rect *rect, struct cobj_matrix *mat);

void gs_get_bezier_rect (struct cobj_rect *rect, struct gs_bezier *bezier);

enum s_status gs_get_string (struct gs_text *text, 
			     struct gs_text_string *string,
			     struct cobj_matrix *mat);
void gs_get_vector (struct gs_text *text, pdf_real_t wx, pdf_real_t wy,
		    struct gs_point *vec);
void gs_get_vector_gc (struct gs_context *gpc, pdf_real_t wx, pdf_real_t wy,
		       struct gs_point *vec);
void gs_string_destroy (struct gs_text_string *string);

void gs_origin_init (struct gs_point *point, struct cobj_rect *rect,
		     enum gs_direction pdir, enum gs_direction sdir);

void gs_origin_half_init (struct gs_point *point, struct cobj_rect *rect,
			  enum gs_direction dir);

gs_edge gs_get_lo_edge (enum gs_direction dir);
gs_edge gs_get_hi_edge (enum gs_direction dir);
gs_cmp gs_get_cmp (enum gs_direction dir);
enum gs_direction gs_dir_neg (enum gs_direction dir);

#endif
