/*
 * PDFlow -- draft.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef draft_h

#define draft_h

enum gr_draft_type {

	GR_D_AREA,
	GR_D_INNER,
	GR_D_OUTER
};

struct gr_draft {
	ref_t refs;

	enum gr_draft_type type;

	pdf_real_t weight;

	pdf_bool_t hin, vin;

	enum gs_direction wdir, ldir;

	struct a_area_geometry g;

	struct gr_region *hreg, *vreg;

	int rows, cols;
	struct gr_draft **cell;

	struct a_area *area;
	
};

enum s_status gr_d__init (struct gr_draft **draft, struct a_area *area,
			  struct args_heur *heur, struct ref_owner *p,
			  const char *fname, int line);
#define gr_d_init(draft, area, heur, p) \
gr_d__init (draft, area, heur, p, __FILE__, __LINE__)

enum s_status gr__d_get_draft (struct gr_draft **draftp, struct a_area *area, 
			       enum gs_direction wdir, enum gs_direction ldir,
			       struct args_heur *heur, struct ref_owner *p, 
			       const char *fname, int line);
#define gr_d_get_draft(draft, area, wdir, ldir, heur, p) \
gr__d_get_draft (draft, area, wdir, ldir, heur, p, __FILE__, __LINE__)

void gr_d__ref (struct gr_draft *draft, struct ref_owner *p,
		const char *fname, int line);
#define gr_d_ref(draf, p) \
gr_d__ref (draft, p, __FILE__, __LINE__)

void gr_d_unref (struct gr_draft *draft, struct ref_owner *p);

typedef enum s_status (*gr_s_draft_f) (struct gr_draft *draft, void *closure);

enum s_status gr_d_traverse (struct gr_draft *draft, 
			     gr_s_draft_f draftf,
			     void *closure);

#endif
