/*
 * PDFlow -- grid.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef grid_h

#define grid_h

#include "grid/region.h"
#include "grid/draft.h"
#include "grid/align.h"
#include "grid/heur.h"

enum gr_logical {

	GR_L_UNDEFINED,

	GR_L_WORD,
	GR_L_LINE,
	GR_L_BOX,
	GR_L_COLUMN,
	GR_L_PAGE

};

#define GR_AGCC_HYPHEN ((agc_code_t) 1)

struct gr_grid;
struct gr_squeeze_param;
struct gr_squeeze_option;
struct gr_line_composer;
struct gr_squeeze_result;
struct gr_th_status;
struct gr_cell;

struct gr_ifc {
	enum gr_logical logical;
	struct a_ifc ifc;

	void (*probe) (struct gr_grid *grid, pdf_real_t pspace,
		       struct args_output *aout,
		       int options, struct gr_squeeze_option *opt);

	enum s_status (*squeeze) (struct gr_grid *grid, 
				  struct agc_context *agc,
				  struct gr_squeeze_option *opt);

	enum s_status (*thread) (struct gr_cell *cell,
				 struct gr_th_status *status);
};

enum gr_page_type {

	GR_PT_COVER,

	GR_PT_SECTION_START,

	GR_PT_SECTION_END,

	GR_PT_ORDINARY,

	GR_PT_EMPTY
};

struct gr_page_info {

	enum gr_page_type type;

	struct gr_margin {
		struct gr_grid *last;

		pdf_real_t pos;

		pdf_real_t space;
	} header, footer;

	struct gr_line *first, *last;

	pdf_real_t graph_space;
};

struct gr_grid {

	ref_t refs;

	enum gs_direction ldir, wdir;

	pdf_real_t weight;

	struct gr_al_alignment al;

	struct gs_point origin;

	int lines, words;

	struct gr_ifc *ifc;

	struct a_area_geometry g;

	/* This is relevant only for page grids */
	struct gr_page_info pinfo;

	pdf_real_t w_graph_space;

	struct a_area *hyphen;

	struct gr_grid_fill {
		struct gr_cell *first;
		int run_len;
		int ridx;
	} fill;

	struct heur_hls_chooser hls;

	struct td_space wspace;

	union {
		struct gr_line *lines;
		struct a_area *area;
	} d;
};

struct gr_cell {
	struct gr_cell *next;
	struct gr_grid *grid;
	struct gr_region reg;
};

struct gr_line {
	struct gr_line *next;
	struct gr_cell *cells;
	struct gr_region reg;
};

typedef enum s_status (*gr_eject) (struct agc_context *agc,
				   struct cobj_rect *bounds,
				   void *closure);


enum s_status gr__init (struct gr_grid **grid, struct gr_grid *parent,
			struct gr_draft *draft, struct args_heur *heur, 
			struct ref_owner *p, const char *fname, int line);
#define gr_init(grid, parent, draft, heur, p) \
gr__init (grid, parent, draft, heur, p, __FILE__, __LINE__)

enum s_status gr__get_grid (struct gr_grid **grid, struct gspg_page *page,
			    struct gr_al_machine *dc,
			    struct args_margins *amys,
			    struct args_heur *heur, struct ref_owner *p, 
			    const char *fname, int line);
#define gr_get_grid(grid, page, glp, amys, heur, p) \
gr__get_grid (grid, page, glp, amys, heur, p, __FILE__, __LINE__)

enum s_status gr_highlight (struct gr_grid *grid, struct agc_context *agc,
			    const char *spec, struct gspg_page *page);

void gr__ref (struct gr_grid *grid, struct ref_owner *p, 
	      const char *fname, int line);
#define gr_ref(grid, p) gr__ref (grid, p, __FILE__, __LINE__)

void gr_unref (struct gr_grid *grid, struct ref_owner *p);

typedef enum s_status (*gr_grid_f) (struct gr_grid *grid, void *closure);
typedef enum s_status (*gr_cell_f) (struct gr_cell *cell, void *closure);

enum s_status gr_traverse (struct gr_grid *grid, gr_grid_f gridf,
			   void *closure);

enum s_status gr_traverse_c (struct gr_grid *grid, gr_cell_f cellf,
			     void *closure);

pdf_bool_t gr_is_word (struct gr_grid *grid);

enum s_status gr_line_append (struct gr_grid *grid,
			      struct gr_line **line);
enum s_status gr_cell_append (struct gr_grid *grid,
			      struct gr_line *line,
			      struct gr_cell *templ,
			      struct gr_cell **cell);
void gr_cell_remove_first (struct gr_grid *grid,
			   struct gr_line *line);



#endif
