/*
 * PDFlow -- heur.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef heur_h

#define heur_h

#include "args.h"
#include "graph/gstate.h"
#include "graph/2d.h"
#include "grid/region.h"
#include "grid/align.h"

#define HEUR_BADNESS_NEW_LINE  1
#define HEUR_BADNESS_NEW_PAGE 10
#define HEUR_BADNESS_LAST_RESORT \
	(HEUR_BADNESS_NEW_PAGE + HEUR_BADNESS_NEW_LINE + 1)

#define HEUR_BL_MAX 10

struct heur_dir_chooser {

	struct args_heur *heur;

	int pstat [GS_D__LAST__], sstat [GS_D__LAST__];
};

struct heur_choice {
	pdf_real_t tval;
	int occurrences;
};


struct heur_bl_chooser {

	struct args_heur *heur;

	struct heur_choice blstat [HEUR_BL_MAX];

	int idx;
	enum gs_direction pdir;

	pdf_bool_t horiz;
};

struct heur_hls_chooser {

	struct heur_choice c;

	pdf_real_t tol, graph;
};


void heur_get_glyph_bounds (struct gf_glyph_info *gi, 
			    struct cobj_rect *bounds,
			    struct gs_point *vec,
			    struct args_heur *heur);

void heur_extend_cols (pdf_real_t lo, pdf_real_t hi, pdf_real_t svc,
		       pdf_real_t *xlo, pdf_real_t *xhi,
		       struct args_heur *heur);

void heur_get_direction (struct gs_text_string *gs, struct gs_text_state *tws,
			 enum gs_direction *pdir, enum gs_direction *sdir);

void heur_dc_init (struct heur_dir_chooser *hc, struct args_heur *heur);
void heur_dc_add (struct heur_dir_chooser *hc, 
		  enum gs_direction pdir, enum gs_direction sdir);
void heur_dc_choose (struct heur_dir_chooser *hc,
		     enum gs_direction *pdir,
		     enum gs_direction *sdir,
		     pdf_bool_t *certain);

void heur_bl_init (struct heur_bl_chooser *hbl, 
		   struct args_heur *heur,
		   enum gs_direction pdir);
void heur_bl_add (struct heur_bl_chooser *hbl, struct gs_point *origin);
void heur_bl_choose (struct heur_bl_chooser *hbl,
		     struct cobj_rect *rect,
		     struct gs_point *origin, 
		     pdf_bool_t *has_origin);

void heur_hls_init (struct heur_hls_chooser *hls, 
		    struct args_heur *heur);
void heur_hls_add (struct heur_hls_chooser *hls, pdf_real_t space);
void heur_hls_choose (struct heur_hls_chooser *hls,
		      pdf_real_t space, pdf_real_t *spu);

pdf_bool_t heur_col_break (struct args_heur *heur, 
			   struct gr_region *preg, 
			   struct gr_region *nreg);
void heur_rotate_to_dir (struct args_heur *heur, 
			 pdf_integer_t rotate, 
			 enum gs_direction *pdir,
			 enum gs_direction *sdir);
#endif
