/*
 * PDFlow -- lc.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef lc_h

#define lc_h

#include "graph/2d.h"
#include "grid/squeeze.h"
#include "grid/align.h"

struct gr_line_cell {

	ref_t refs;

	struct gr_line_cell *next;

	struct agc_context *agc;

	struct cobj_rect rect, bounds;

	struct gs_point origin;

	struct gr_grid *grid;

	struct td_space wspace;
};

struct gr_line_rects {

	struct gr_al_alignment al;

	struct cobj_rect lavail;

	pdf_real_t line_size;

	pdf_real_t strict, blanks, last_word_blank;

	pdf_real_t dhead, dtail, ratio;

	pdf_real_t ascent, descent;

};

struct gr_lc_param {

	struct args_output *aout;

	struct args_heur *heur;
};

struct gr_line_composer {

	ref_t refs;

	gr_eject eject;

	pdf_bool_t dirty;

	void *ejclosure;

	struct gr_lc_param lcp;

	struct gr_line_rects r;

	pdf_real_t badness;

	struct gr_line_cell *cells;

	struct td_pool *tdp;

	struct td_float *tdf;

	pdf_bool_t page_flushed;
};

enum s_status gr_lc__new (struct gr_line_composer **lc, gr_eject eject,
			  void *ejclosure, struct gr_lc_param *lcp,
			  struct cobj_rect *rect, 
			  struct ref_owner *p, const char *fname, int line);
#define gr_lc_new(lc, eject, ejclosure, lcp, rect, p) \
gr_lc__new (lc, eject, ejclosure, lcp, rect, p, __FILE__, __LINE__)

void gr_lc__ref (struct gr_line_composer *lc, struct ref_owner *p,
		 const char *fname, int line);
#define gr_lc_ref(lc, p) gr_lc__ref (lc, p, __FILE__, __LINE__)
void gr_lc_unref (struct gr_line_composer *lc, struct ref_owner *p);

enum s_status gr_lc_feed (struct gr_line_composer *lc, struct gr_grid *grid);

enum s_status gr_lc_finalize (struct gr_line_composer *lc);

#endif
