/*
 * PDFlow -- sig.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

/*
 * This module shall implement the affinity code. Affinity is used
 * by grid/ modules to tell how much two lines look "similar".
 * If they do, then they can be merged into a single paragraph.
 * In addition, affinity will be used to build a paragraph "library"
 * to recognize patterns like:
 *	- How much line space to place before/after a paragraph
 *	- Does the paragraph always start a new page?
 *	- Does the paragraph always follow another type of paragraph
 *		(useful to avoid orphans & widows).
 *
 */

#ifndef sig_h

#define sig_h

#include "grid/grid.h"

/*
 * The signature structure shall be fixed in size, without pointers
 * and other stuff. It can be statically allocated by client apps,
 * but it is completely opaque (this algorithm is likely to change much)
 */

struct gr_sig_signature {

	/* Average line size */
	pdf_real_t size;

};

/* Init a signature, given a line grid */
void gr_sig_mksig (struct gr_sig_signature *sig, 
		   struct gr_grid *grid);

/* Obtain affinity. Affinity ranges between 0 and 1 */
pdf_real_t gr_sig_affinity (struct gr_sig_signature *sig1,
			    struct gr_sig_signature *sig2,
			    struct args_heur *heur);

/* Tell whether two adjacent lines can be merged */
pdf_real_t gr_sig_can_merge (struct gr_sig_signature *sig1,
			     struct gr_sig_signature *sig2,
			     struct args_heur *heur);

#endif
