/*
 * PDFlow -- thread.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef thread_h

#define thread_h

#include "graph/2d.h"
#include "grid/heur.h"
#include "grid/grid.h"

enum gr_th_zone {

	GR_TH_Z_HEADER,
	GR_TH_Z_BODY,
	GR_TH_Z_FOOTER

};

/* Info available to gr_th_grid_f when traversing a grid */
struct gr_th_info {

	/* This grid's parent */
	struct gr_grid *parent;

	/* The "column" this grid belongs to. May not be an
	 * actual Column grid. For instance it may be an orphan line */
	struct gr_grid *column;

	/* The spacing between this line and the previous one */
	struct td_space lspace;

	/* The spacing between this word and the next one */
	struct td_space wspace;

	/* Whether this is this parent's last cell in this row */
	pdf_bool_t last;

	/* Cell this grid belongs to */
	struct gr_cell *cell;
};

typedef enum s_status (*gr_th_grid_f) (struct gr_grid *grid, 
				       struct gr_th_info *info,
				       void *closure);

struct gr_th_status {

	gr_th_grid_f gridf;

	void *closure;

	struct args_heur *heur;

	struct gr_grid *last_line;

	struct gr_th_info info;

	pdf_bool_t graph_allowed;

	enum gr_th_zone zone;

	struct gr_page_info *pinfo;
};

enum s_status gr_th_page_thread (struct gr_cell *cell,
				 struct gr_th_status *ths);

enum s_status gr_th_word_thread (struct gr_cell *cell,
				 struct gr_th_status *ths);

enum s_status gr_th_box_thread (struct gr_cell *cell,
				struct gr_th_status *ths);

enum s_status gr_th_column_thread (struct gr_cell *cell,
				   struct gr_th_status *ths);

enum s_status gr_th_line_thread (struct gr_cell *cell,
				 struct gr_th_status *ths);

enum s_status gr_th_traverse (struct gr_grid *grid, struct args_heur *heur,
			      gr_th_grid_f gridf, void *closure);


#endif
