/*
 * PDFlow -- pdfw.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef pdfw_h

#define pdfw_h

#include "types/cobj.h"

#define PDFW_ARG_MAX 10

enum pdfw_opcode_type {

	PDFW_OP_SET_CTM,
	PDFW_OP_LINE_WIDTH,
	PDFW_OP_LINE_CAP,
	PDFW_OP_LINE_JOIN,
	PDFW_OP_MITER_LIMIT,
	PDFW_OP_DASH_PATTERN,
	PDFW_OP_RENDERING_INTENT,

	PDFW_OP_FONT,
	PDFW_OP_TEXT_MATRIX,
	PDFW_OP_CHAR_SPACE,
	PDFW_OP_WORD_SPACE,
	PDFW_OP_SCALE,
	PDFW_OP_LEADING,
	PDFW_OP_RENDER,
	PDFW_OP_RISE,

	PDFW_OP_GS,

	PDFW_OP_BT,
	PDFW_OP_ET,

	PDFW_OP_DRAW_LINE,
	PDFW_OP_DRAW_BEZIER,

	PDFW_OP_DRAW_TEXT,

	PDFW_OP_CLOSE_PATH,

	PDFW_OP_SHADING,
	PDFW_OP_XOBJECT,
	PDFW_OP_INLINE_IMAGE,

	PDFW_OP_SET_GRAY,
	PDFW_OP_SET_RGB,
	PDFW_OP_SET_CMYK,

	PDFW_OP_SET_CSPACE,
	PDFW_OP_SET_COLOUR,

	PDFW_OP_PUSH,
	PDFW_OP_POP,

	PDFW_OP_ADD_CLIPPING_PATH
};

struct pdfw_writer;

struct pdfw_opcode {

	struct pdfw_opcode *next;

	enum pdfw_opcode_type type;

	struct cobj_matrix matrix;

	struct xref_obj_array array;

	int args;
	struct obj arg [PDFW_ARG_MAX];

	struct pdfw_opcode_desc *opd;
};

enum s_status pdfw_create (struct f_file *dest,
			   struct v_selection *sel,
			   struct pdfw_writer **pdfw);
void pdfw_destroy (struct pdfw_writer *pdfw);

enum s_status pdfw_append (struct pdfw_writer *pdfw,
			   enum pdfw_opcode_type type, ...);

enum s_status pdfw_flush (struct pdfw_writer *pdfw);

enum s_status pdfw_ll_flush (struct pdfw_writer *pdfw);

#endif
