/*
 * PDFlow -- refs.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef refs_h

#define refs_h

#ifdef DEBUG_REFS

#include <assert.h>
enum ref_owner_type {
	REF_OT_ENDPOINT,
	REF_OT_REF
};

struct ref_owner {
	enum ref_owner_type type;
	const char *name;
};

struct ref_ptr {
	struct ref_owner *p;
	const char *fname;
	int line;
};

typedef struct ref {
	
	struct ref *next;

	int n, cap;

	struct ref_owner myself;

	struct ref_ptr *table;

} ref_t;
void ref__init (ref_t *r, const char *name, struct ref_owner *p,
	       const char *fname, int line);
void ref_zero (ref_t *r);
void ref_shutdown (void);
void ref__up (ref_t *r, struct ref_owner *p, const char *fname, int line);
int ref_val (ref_t *t);
int ref_down (ref_t *r, struct ref_owner *p);

void ref_init_owner (struct ref_owner *p, const char *name);

struct ref_owner *ref_rown (ref_t *r);

#else
typedef int ref_t;
struct ref_owner {
	/* empty */
};

#define ref__init(r, name, p, file, line) (*r) = 1
#define ref_destroy(r) 
#define ref_val(r) (*r)
#define ref__up(r, p, file, line) (*r)++
#define ref_down(r, p) (--(*r))
#define ref_zero(r)

#define ref_shutdown()

#define ref_init_owner(p, name)

#define ref_rown(r) NULL

#endif


#define ref_init(r, name, p) ref__init (r, name, p, __FILE__, __LINE__)
#define ref_up(r, p) ref__up (r, p, __FILE__, __LINE__)

#endif
