/*
 * PDFlow -- cmap.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef cmap_h

#define cmap_h

#include "refs.h"
#include "fmt/fields.h"
#include "types/types.h"

#define CMAP_SYSTEM_INFO_I(name, field, major, minor, mand) \
XREF_FUNCTION_I (name, struct cmap_system_info, field, \
                 &cmap_system_info_pstable, major, minor, mand)

enum cmap_wmode {
	CMAP_WM_HORIZONTAL = 0,
	CMAP_WM_VERTICAL
};

struct cmap_table {
	struct cmap_table *next;

	uint8_t lower [4], upper [4];
	size_t size, blen;
	uint32_t *code;
};

struct cmap_system_info {

	struct t_type t;

#define CMAP_SYSTEM_INFO
#include "cmap.tab.h"	

};

struct cmap {

	struct t_type t;

	struct cmap *base;

	struct cmap_table *tables;

#define CMAP
#include "cmap.tab.h"	

};

T_PROTOTYPES (cmap, cmap);
T_PROTOTYPES (cmap_system_info, cmap_system_info);

enum s_status cmap__build_identity_h (struct cmap **cmap, struct xref_db *rdb,
				      struct ref_owner *p, const char *fname,
				      int line);
enum s_status cmap__build_identity_v (struct cmap **cmap, struct xref_db *rdb,
				      struct ref_owner *p, const char *fname,
				      int line);
#define cmap_build_identity_h(cmap, rdb, p) \
cmap__build_identity_h (cmap, rdb, p, __FILE__, __LINE__)
#define cmap_build_identity_v(cmap, rdb, p) \
cmap__build_identity_v (cmap, rdb, p, __FILE__, __LINE__)

uint32_t cmap_get_cid (struct cmap *cmap, struct obj_bstr *s, size_t *len);

enum s_status cmap_create_table (struct cmap_table **table,
				 struct obj *start, struct obj *end);
enum s_status cmap_dup_table (struct cmap_table *stable,
			      struct cmap_table **dtable);
void cmap_destroy_table (struct cmap_table *table);

uint32_t *cmap_find_slot (struct cmap_table *table, 
			  struct obj_bstr *first, struct obj_bstr *last);

#endif
