/*
 * PDFlow -- cobj.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef cobj_h

#define cobj_h

#include <sys/time.h>
#include "fmt/objbr.h"
#include "fmt/fields.h"

struct cobj_enum_desc {
	int value;
	const char *name;
	int major, minor;
	void *data;
};

enum cobj_rect_overlap {
	COBJ_RO_NONE,
	COBJ_RO_OVERLAP,
	COBJ_RO_INSIDE,
	COBJ_RO_OUTSIDE,
	COBJ_RO_SAME
};

#define COBJ_ENUM_DECLARE(name, ttype) \
struct name {				\
	int type;			\
	ttype t;			\
	struct obj unkval;		\
}

COBJ_ENUM_DECLARE (cobj_enum, void *);

struct cobj_enum_array {
	size_t n;
	struct cobj_enum *data;
};

COBJ_ENUM_DECLARE (cobj_enum_obj, struct obj);

struct cobj_enum_obj_array {
	size_t n;
	struct cobj_enum_obj *data;
};

struct cobj_date {

	time_t time;

	int tz_hour;

	int tz_minutes;
};

struct cobj_rect {

	pdf_real_t left;

	pdf_real_t top;

	pdf_real_t right;

	pdf_real_t bottom;

};

struct cobj_matrix {
	pdf_real_t a [3][3];
};

struct cobj_id {
	struct obj_bstr id1, id2;
};

extern struct cobj_matrix cobj_eye;

extern struct xref_pstable cobj_enum_pstable;
extern struct xref_pstable cobj_date_pstable;
extern struct xref_pstable cobj_rect_pstable;
extern struct xref_pstable cobj_matrix_pstable;
extern struct xref_pstable cobj_matrix9_pstable;
extern struct xref_pstable cobj_id_pstable;
#define COBJ_ENUM(name, field, desc, major, minor, mand) \
XREF_FUNCTION_I_DATA (name, struct cobj_enum, field, \
                      &cobj_enum_pstable, desc, major, minor, mand)
#define COBJ_ENUM_OBJ(name, field, desc, major, minor, mand) \
XREF_FUNCTION_I_DATA (name, struct cobj_enum_obj, field, \
                      &cobj_enum_pstable, desc, major, minor, mand)
#define COBJ_ENUM_S_ARRAY(name, field, desc, major, minor, mand) \
XREF_S_ARRAY_DATA (name, struct cobj_enum_array, field, \
                   &cobj_enum_pstable, desc, major, minor, mand)
#define COBJ_ENUM_A_ARRAY(name, field, desc, major, minor, mand) \
XREF_A_ARRAY_DATA (name, struct cobj_enum_array, field, \
                   &cobj_enum_pstable, desc, major, minor, mand)
#define COBJ_DATE(name, field, major, minor, mand) \
XREF_FUNCTION_I (name, struct cobj_date, field, 			 \
                 &cobj_date_pstable, major, minor, mand)
#define COBJ_RECT(name, field, major, minor, mand) \
XREF_FUNCTION_I (name, struct cobj_rect, field, 			 \
                 &cobj_rect_pstable, major, minor, mand)
#define COBJ_MATRIX(name, field, major, minor, mand) \
XREF_FUNCTION_I (name, struct cobj_matrix, field, 			 \
		 &cobj_matrix_pstable, major, minor, mand)
#define COBJ_MATRIX9(name, field, major, minor, mand) \
XREF_FUNCTION_I (name, struct cobj_matrix, field, 			 \
		 &cobj_matrix9_pstable, major, minor, mand)
#define COBJ_ID(name, field, major, minor, mand) \
XREF_FUNCTION_I (name, struct cobj_id, field, \
		 &cobj_id_pstable, major, minor, mand)

enum s_status cobj_id_cross (struct cobj_id *id, struct cobj_id *sid,
			     struct ob_bridge *ob);

void cobj_date_now (struct cobj_date *date);

void cobj_rect_set_null (struct cobj_rect *rect);
int cobj_rect_is_null (struct cobj_rect *rect);

enum cobj_rect_overlap cobj_rect_overlaps (struct cobj_rect *rect1, 
					   struct cobj_rect *rect2);
void cobj_rect_union (struct cobj_rect *drect, struct cobj_rect *srect);
void cobj_rect_intersect (struct cobj_rect *drect, struct cobj_rect *srect);
void cobj_rect_sub (struct cobj_rect *drect, struct cobj_rect *srect);
void cobj_rect_add_point (struct cobj_rect *drect, pdf_real_t x, 
			  pdf_real_t y);

pdf_real_t cobj_rect_width (struct cobj_rect *rect);
pdf_real_t cobj_rect_height (struct cobj_rect *rect);

void cobj_matrix_set_null (struct cobj_matrix *mat);
int cobj_matrix_is_null (struct cobj_matrix *mat);

void cobj_matrix_mul (struct cobj_matrix *a, struct cobj_matrix *b,
		      struct cobj_matrix *ans, int red);
void cobj_matrix_inv (struct cobj_matrix *a, struct cobj_matrix *ans, int red);
void cobj_matrix_div (struct cobj_matrix *a, struct cobj_matrix *b,
		      struct cobj_matrix *ans, int red);

void cobj_translate (struct cobj_rect *rect, struct cobj_matrix *mat);

#endif
