/*
 * PDFlow -- colour.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef colour_h

#define colour_h

#include "refs.h"

#include "types/function.h"
#include "types/resdict.h"
#include "types/types.h"

struct rd_resdict;
struct pt_pattern;

enum gc_colour_space_family {
	GC_CSF_NULL = 0,

	GC_CSF_DEVICE_GRAY,
	GC_CSF_DEVICE_RGB,
	GC_CSF_DEVICE_CMYK,

	GC_CSF_CAL_GRAY,
	GC_CSF_CAL_RGB,
	GC_CSF_LAB,
	GC_CSF_ICC_BASED,

	GC_CSF_PATTERN,
	GC_CSF_INDEXED,
	GC_CSF_SEPARATION,
	GC_CSF_DEVICE_N,

	GC_CSF_OTHER
};

/* Need to define these early, since these objects are recursive */
T_PROTOTYPES (gc_cspace, gc_colour_space);

#define GC_CSPACE_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct gc_colour_space, field, \
                 &gc_cspace_pstable, major, minor, mand)

#define GC_LOOKUP(name, field, mand) \
XREF_FUNCTION_I (name, struct obj_bstr, field, \
                 &gc_lookup_pstable, 1, 0, mand)


enum gc_device_n_type {
	GC_DNT_DEVICE_N,
	GC_DNT_NCHANNEL
};

struct gc_d_cie {
	pdf_real_t wp [3];
	pdf_real_t bp [3];
	pdf_real_t gamma [3];
	struct cobj_matrix matrix;
	pdf_real_t amin, amax;
	pdf_real_t bmin, bmax;
};

struct gc_d_icc_based {
	int n;
	pdf_real_t min [4], max [4];
	struct obj metadata;
	struct obj_region data;
	struct gc_colour_space *alternate; 
};

struct gc_d_pattern {

#define GC_D_PATTERN
#include "types/colour.tab.h"	

};

struct gc_d_indexed {

#define GC_D_INDEXED
#include "types/colour.tab.h"	

};

struct gc_d_separation {

#define GC_D_SEPARATION
#include "types/colour.tab.h"

};

struct gc_d_device_n {

	enum gc_device_n_type type;

	struct gc_d_device_n_col {
		char *name;

		struct gc_colour_space *sep;

		int pidx;

		struct fun_function *dot_gain;
		pdf_real_t solidity;
		pdf_integer_t printing_order;
	} *col;

	struct fun_function *def_dg;

	struct fun_function *tint_transform;

	struct gc_colour_space *process_space;

	struct gc_colour_space *alternate_space;

	int n;

};

union gc_cspace_data {
	struct gc_d_cie cie;
	struct gc_d_icc_based icc;
	struct gc_d_pattern pattern;
	struct gc_d_indexed indexed;
	struct gc_d_separation separation;
	struct gc_d_device_n device_n;
};

struct gc_colour_space {

	struct t_type t;
	
	struct gc_colour_space_ifc *ifc;

#define GC_COLOUR_SPACE
#include "types/colour.tab.h"	

};

struct gc_f_device_gray {
	pdf_real_t gray;
};

struct gc_f_device_rgb {
	pdf_real_t red;
	pdf_real_t green;
	pdf_real_t blue;
};

struct gc_f_device_cmyk {
	pdf_real_t cyan;
	pdf_real_t magenta;
	pdf_real_t yellow;
	pdf_real_t black;
};

struct gc_f_cal_gray {
	pdf_real_t a;
};

struct gc_f_cal_rgb {
	pdf_real_t a;
	pdf_real_t b;
	pdf_real_t c;
};

struct gc_f_lab {
	pdf_real_t l;
	pdf_real_t a;
	pdf_real_t b;
};

struct gc_f_icc_based {
	pdf_real_t v [4];
};

struct gc_f_indexed {
	pdf_real_t index;
};

struct gc_f_separation {
	pdf_real_t value;
};

struct gc_f_device_n {
	pdf_real_t *value;
};

struct gc_colour_state {

	ref_t refs;

	struct gc_colour_space *sp, *defsp;

	struct pt_pattern *pattern;

	union {
		struct gc_f_device_gray device_gray;
		struct gc_f_device_rgb device_rgb;
		struct gc_f_device_cmyk device_cmyk;
		struct gc_f_cal_gray cal_gray;
		struct gc_f_cal_rgb cal_rgb;
		struct gc_f_lab lab;
		struct gc_f_icc_based icc;
		struct gc_f_indexed indexed;
		struct gc_f_separation separation;
		struct gc_f_device_n device_n;
	} u;
};

int gc_is_pattern (struct gc_colour_state *cst);

enum s_status gc_set_colour_state (struct obj *name, 
				   struct xref_db *rdb,
				   struct rd_resdict *rd,
				   struct gc_colour_state **cst,
				   struct ref_owner *p);
enum s_status gc_select_colour (struct gc_colour_state **cst,
				struct pt_pattern *pattern, 
				pdf_real_t *val,
				struct ref_owner *p);
enum s_status gc_get_colour (struct gc_colour_state *cst,
			     struct pt_pattern **pattern, 
			     pdf_real_t **val, int *components,
			     struct ref_owner *p);

int gc_cspace_equal (const struct gc_colour_space *csp1, 
		     const struct gc_colour_space *csp2);

int gc_cstate_equal (const struct gc_colour_state *cst1,
		     const struct gc_colour_state *cst2);

enum gc_colour_space_family gc_get_family (struct gc_colour_space *csp);

int gc_get_components (struct gc_colour_state *cst);

enum s_status gc_cstate__init (struct gc_colour_state **cst,
			       struct xref_db *rdb,
			       struct ref_owner *p,
			       const char *fname, int line);
#define gc_cstate_init(cst, rdb, p) \
gc_cstate__init (cst, rdb, p, __FILE__, __LINE__)
void gc_cstate__ref (struct gc_colour_state *cst, 
		     struct ref_owner *p, const char *fname, int line);
#define gc_cstate_ref(cst, p) gc_cstate__ref (cst, p, __FILE__, __LINE__)


void gc_cstate_unref (struct gc_colour_state *cst, struct ref_owner *p);
#endif
