/*
 * PDFlow -- filespec.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef filespec_h

#define filespec_h

#include "refs.h"
#include "types/types.h"
#include "types/cobj.h"
#include "types/txtstring.h"
#include "fmt/fields.h"

enum fspec_type {
	FS_FT_DEFAULT,
	FS_FT_URI,
};

enum fspec_filesys {
	FS_FT_UF,
	FS_FT_F,

	FS_FT_DOS,
	FS_FT_MAC,
	FS_FT_UNIX,
};

struct fspec_mac {
#define FSPEC_MAC
#include "types/filespec.tab.h"
};

struct fspec_params {
#define FSPEC_PARAMS
#include "types/filespec.tab.h"
};

struct fspec_embedded {
#define FSPEC_EMBEDDED
#include "types/filespec.tab.h"
};

struct fspec_ef {
#define FSPEC_EF
#include "types/filespec.tab.h"
};

struct fspec_rf {
#define FSPEC_RF
#include "types/filespec.tab.h"

};

struct fspec_info {

	char *subtype;

	size_t size;
	
	struct cobj_date creation_date;

	struct cobj_date mod_date;

	struct fspec_mac mac;

};

struct fspec_related_info {

	char *fname;

	struct f_file *file;

	struct fspec_info info;
};

struct fspec_specification {

	struct t_type t;

#define FSPEC_DICTIONARY
#include "types/filespec.tab.h"

	struct obj ufo;

	struct f_file *file;
	
	char *filename;

	enum fspec_type type;

	enum fspec_filesys fsys;

	struct obj_bstr descname;
};

T_PROTOTYPES (fspec_fs, fspec_specification);

enum s_status fspec_get (struct fspec_specification *fs, 
			 struct f_file **f, struct fspec_info *info,
			 struct fspec_related_info **rinfo, int *relno);

void fspec_info_destroy (struct fspec_info *info);
void fspec_related_info_destroy (struct fspec_related_info *info, int relno);

#endif
