/*
 * PDFlow -- font.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef font_h

#define font_h

#include "types/types.h"
#include "types/cobj.h"
#include "types/resdict.h"
#include "types/cmap.h"

#define GF_FONT_DESC_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct gf_font_desc, field, \
                 &gf_desc_pstable, major, minor, mand)

#define GF_FONT_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct gf_font, field, \
                 &gf_pstable, major, minor, mand)

#define GF_ENCODING_ENUM(name, field, desc, major, minor, mand) \
XREF_FUNCTION_I_DATA (name, struct gf_encoding_enum, field, \
		      &cobj_enum_pstable, desc, major, minor, mand)

enum gf_font_type {

	GF_FT_TYPE0,
	GF_FT_TYPE1,
	GF_FT_MMTYPE1,
	GF_FT_TYPE3,
	GF_FT_TRUE_TYPE,
	GF_FT_CID_FONT_TYPE0,
	GF_FT_CID_FONT_TYPE2

};

enum gf_cid_to_gid {

	GF_C2G_IDENTITY,
	GF_C2G_STREAM

};

struct gf_font_array {
	size_t n;
	struct gf_font **data;
};

struct gf_glyph_info {

	struct cobj_rect rect;

	pdf_real_t wx, wy;

	int size;

	uint32_t unicode;
};

struct gf_table {

	pdf_integer_t first, size;

	struct gf_glyph_info *info;

	struct gf_glyph_info unmapped;

};

struct gf_encoding {

	struct t_type t;

#define GF_ENCODING
#include "types/font.tab.h"
};

COBJ_ENUM_DECLARE (gf_encoding_enum, struct gf_encoding *);

struct gf_font_desc {

	struct t_type t;

#define GF_FONT_DESC
#include "types/font.tab.h"
};

struct gf_simple_font {

	struct gf_table table;

#define GF_SIMPLE_FONT
#include "types/font.tab.h"
};

struct gf_t0_entry {

	struct gf_glyph_info info;

	uint32_t cid;

};

struct gf_type0_font {
	
	struct cmap *cmap;

	pdf_integer_t size;

	struct gf_t0_entry *t0e, def_entry;
	
	struct obj_region *c2g;

#define GF_TYPE0_FONT
#include "types/font.tab.h"
};

struct gf_type3_font {

	struct gf_table table;

#define GF_TYPE3_FONT
#include "types/font.tab.h"
};

struct gf_cid_font {
#define GF_CID_FONT
#include "types/font.tab.h"
};

union gf_data {
	struct gf_type0_font type0;
	struct gf_simple_font simple;
	struct gf_type3_font type3;
	struct gf_cid_font cid;
};

struct gf_font {

	struct t_type t;

	pdf_real_t space_width, leading, ascent, descent;

	struct gf_font_ifc *ifc;

	enum cmap_wmode wmode;

#define GF_FONT
#include "types/font.tab.h"
};

T_PROTOTYPES (gf, gf_font);
T_PROTOTYPES (gf_desc, gf_font_desc);
T_PROTOTYPES (gf_encoding, gf_encoding);

struct gf_glyph_info *gf_get_glyph_info (struct gf_font *font,
					 uint8_t *string, size_t limit);

#endif

