/*
 * PDFlow -- function.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef function_h

#define function_h

#include "types/types.h"
#include "types/cobj.h"
#include "hash.h"

T_PROTOTYPES (fun, fun_function);

enum fun_type {
	FUN_T_SAMPLED = 0,
	FUN_T_EXPONENTIAL = 2,
	FUN_T_STITCHING = 3,
	FUN_T_POSTSCRIPT = 4
};

struct fun_bound {
	pdf_real_t lower, upper;
};

struct fun_region {
	pdf_integer_t m;
	struct fun_bound *bound;
};

extern struct xref_pstable fun_region_pstable;
#define FUN_REGION(name, field, major, minor, mand) \
XREF_FUNCTION_I (name, struct fun_region, field, \
                 &fun_region_pstable, major, minor, mand)

#define FUN_FUNCTION_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct fun_function, field, \
                 &fun_pstable, major, minor, mand)
#define FUN_ENUM(name, field, desc, major, minor, mand) \
XREF_FUNCTION_I_DATA (name, struct fun_enum, field, \
                      &cobj_enum_pstable, desc, major, minor, mand)

struct fun_sampled {
	size_t scount;
	pdf_integer_t *samples;
#define FUN_SAMPLED
#include "types/function.tab.h"
};

struct fun_exponential {
#define FUN_EXPONENTIAL
#include "types/function.tab.h"
};

struct fun_stitching {
#define FUN_STITCHING
#include "types/function.tab.h"
};

struct fun_postscript {
	struct f_file *file;
	struct psi_interpreter *pi;
};

union fun_data {
	struct fun_sampled sampled;
	struct fun_exponential exponential;
	struct fun_stitching stitching;
	struct fun_postscript postscript;
};

struct fun_function {

	struct t_type t;

	struct fun_ifc *ifc;  

#define FUN_FUNCTION
#include "types/function.tab.h"
};

struct fun_request {
	pdf_real_t *in;
	pdf_real_t *out;
};

COBJ_ENUM_DECLARE (fun_enum, struct fun_function *);

struct fun_enum_array {
	size_t n;
	struct fun_enum *data;
};


enum s_status fun_exec (struct fun_function *func, 
			struct fun_request *request);

enum s_status fun_get_checksum (struct fun_function *func, struct hash *cks);

#endif
