/*
 * PDFlow -- page.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef page_h

#define page_h

#include "refs.h"
#include "types/types.h"
#include "types/cobj.h"
#include "types/resdict.h"
#include "btree.h"

struct pg_inheritable {

#define PG_INHERITABLE
#include "page.tab.h"

};

struct pg_node {

	struct t_type t;

	struct pg_inheritable inh;

#define PG_NODE
#include "page.tab.h"

	struct pg_node *next;
};

struct pg_page {

	struct t_type t;

	struct pg_inheritable inh;

#define PG_PAGE
#include "page.tab.h"
};

struct pg_pages {

	struct t_type t;

	struct bt_tree *tree;

	struct pg_node *delnodes;
};

struct pg_contents {

	int elements;

	struct f_file **file;
};

T_PROTOTYPES (pg_node, pg_node);
T_PROTOTYPES (pg_page, pg_page);
T_PROTOTYPES (pg_pages, pg_pages);

#define PG_PAGES_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct pg_pages, field, \
                 &pg_pages_pstable, major, minor, mand)

enum s_status pg_page_insert (struct pg_page **pagep, 
			      struct pg_pages *pages,
			      struct bt_cursor *pc);
struct pg_page *pg_get (struct bt_cursor *pc);

enum s_status pg_page_get_contents (struct pg_pages *pages,
				    struct pg_page *page,
				    struct pg_contents *pgc);
void pg_page_destroy_contents (struct pg_contents *pgc);

void pg_page_set_resdict (struct pg_page *page,
			  struct rd_resdict *rd);

enum s_status pg_page_xfer_contents (struct pg_page *page,
				     struct obj_region *reg,
				     struct xref_bind *robj,
				     struct v_selection *sel);
enum s_status pg_page_store_contents (struct pg_page *page,
				      struct obj_region *reg,
				      struct xref_bind *robj,
				      struct v_selection *sel);
pdf_integer_t pg_get_rotate (struct pg_page *page);

#endif
