/*
 * PDFlow -- shading.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef shading_h

#define shading_h

#include "refs.h"
#include "fmt/obj.h"
#include "types/function.h"
#include "types/cobj.h"
#include "types/colour.h"

enum shd_shading_type {
	SHD_ST_NULL = 0,

	SHD_ST_FUNCTION_BASED = 1,
	SHD_ST_AXIAL = 2,
	SHD_ST_RADIAL = 3,
	SHD_ST_FREE_GOURAUD = 4,
	SHD_ST_LATTICE_GOURAUD = 5,
	SHD_ST_COONS_PATCH = 6,
	SHD_ST_TENSOR_PATCH = 7
};

struct shd_s_functions {

	int elements;
	struct f_function **function;
};

struct shd_s_function_based {
#define SHD_FUNCTION_BASED
#include "shading.tab.h"	
};

struct shd_s_axial {

#define SHD_AXIAL
#include "shading.tab.h"	
};

struct shd_s_radial {

#define SHD_RADIAL
#include "shading.tab.h"	
};

struct shd_s_free_gouraud {

#define SHD_FREE_GOURAUD
#include "shading.tab.h"	

};

struct shd_s_lattice_gouraud {

#define SHD_LATTICE_GOURAUD
#include "shading.tab.h"	
};

struct shd_s_coons_patch {

#define SHD_COONS_PATCH
#include "shading.tab.h"	
};

struct shd_s_tensor_patch {

#define SHD_TENSOR_PATCH
#include "shading.tab.h"	
};

union shd_data {

	struct shd_s_function_based function_based;
	struct shd_s_axial axial;
	struct shd_s_radial radial;
	struct shd_s_free_gouraud free_gouraud;
	struct shd_s_lattice_gouraud lattice_gouraud;
	struct shd_s_coons_patch coons_patch;
	struct shd_s_tensor_patch tensor_patch;
};


struct shd_shading {

	struct t_type t;

	struct shd_shading_ifc *ifc;

#define SHD_SHADING
#include "shading.tab.h"

};

T_PROTOTYPES (shd, shd_shading);

#define SHD_FIELD_R(name, field, major, minor, mand) \
XREF_FUNCTION_R (name, struct shd_shading, field, \
                 &shd_pstable, major, minor, mand)



#endif
