/*
 * PDFlow -- txtstring.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef txtstring_h

#define txtstring_h

#include "refs.h"
#include "hash.h"
#include "fmt/fields.h"
#include "fmt/objbr.h"

extern struct xref_pstable ts_pstable;

struct ts_string {

	struct obj obj;

	enum ts_encoding {
		TS_E_PDF_DOC,
		TS_E_UTF16BE
	} encoding;

	int tokens;
	struct ts_token {
		char lang [2];
		char country [2];
		char *s;
		size_t len;
	} *token;

};

enum s_status ts_init (struct ts_string *ts, struct xref_db *rdb,
		       struct obj *src);
enum s_status ts_init_c (struct ts_string *ts, const char *s);
enum s_status ts_update (struct ts_string *ts, struct hash_status *hs);
void ts_destroy (struct ts_string *ts);
enum s_status ts_parser (struct xref_db *rdb, struct xref_bind *robj,
			 struct ref_owner *p, struct obj *src,
			 struct xref_field *fd, const char *fname,
			 int line);
enum s_status ts_ser (struct xref_db *rdb, struct xref_bind *robj,
		      struct v_selection *sel, struct obj *obj,
		      enum xref_changed *changed, struct xref_field *fd);

enum s_status ts_cross (struct ts_string *ts, struct ts_string *sts,
			struct ob_bridge *ob);

#define TS_FIELD_I(name, field, major, minor, mand) \
XREF_FUNCTION_I (name, struct ts_string, field, \
                 &ts_pstable, major, minor, mand)
#define TS_FIELD_S_ARRAY(name, field, major, minor, mand) \
XREF_A_ARRAY (name, struct ts_string, field, \
              &ts_pstable, major, minor, mand)

enum s_status ts_strip_lang (struct ts_string *src, 
			     struct obj *dest);

int ts_is_null (struct ts_string *ts); 

#endif
