/*
 * PDFlow -- types.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef types_h

#define types_h

#include "hash.h"
#include "fmt/fields.h"
#include "fmt/objbr.h"
#include "refs.h"

#ifndef T_CAST_T
#define T_CAST_T(t) (&(t) -> t)
#endif

struct t_tag {
	pdf_integer_t rid;
	struct obj_reference ref;
	struct obj_bstr name;
};

struct t_type {
	ref_t refs;
	struct xref_db *rdb;
	struct obj obj;
	struct t_tag tag;
};

typedef enum s_status (*t_ini) (struct t_type *t, struct xref_db *rdb,
				struct xref_bind *robj, struct ref_owner *p);
typedef void (*t_fini) (struct t_type *t, struct ref_owner *p);

struct t_cache_closure {
	struct xref_field *fd;
	t_fini fini;
};

#define T_PROTOTYPES(pfx, type) \
extern const char *pfx##_cache_xkey; 					\
extern struct xref_pstable pfx##_pstable; 				\
extern struct t_cache_closure pfx##_cache_closure; 			\
struct pfx##_array { size_t n; struct type **data; };			\
enum s_status pfx##_init (struct type **t, struct xref_db *rdb,      	\
			  struct xref_bind *robj, const struct obj *obj,\
			  struct ref_owner *p, const char *fname,       \
                          int line);					\
enum s_status pfx##_new (struct type **t, struct xref_db *rdb,		\
			 struct ref_owner *p, const char *fname, 	\
			 int line);					\
enum s_status pfx##_cross (struct type **t, struct type *st, 		\
			   struct xref_bind *robj, struct ob_bridge *ob,\
 			   struct ref_owner *p, const char *fname, 	\
			   int line);					\
enum s_status pfx##_cache_init (struct xref_db *rdb);			\
void pfx##_cache_destroy (struct xref_db *rdb);				\
void pfx##_ref (struct type *t, struct ref_owner *p,			\
                const char *fname, int line);				\
void pfx##_unref (struct type *t, struct ref_owner *p);			\
void pfx##_funref (struct ref_owner *p,	 struct xref_field *fd);	\
enum s_status pfx##_parser (struct xref_db *rdb, struct xref_bind *robj,\
			    struct ref_owner *p,			\
 			    struct obj *src, struct xref_field *fd,     \
			    const char *fname, int line);		\
enum s_status pfx##_ser (struct xref_db *rdb, struct xref_bind *robj,	\
			 struct v_selection *sel, struct obj *obj, 	\
			 enum xref_changed *changed, struct xref_field *fd);\
enum s_status pfx##_sync (struct xref_db *rdb, struct xref_bind *robj,	\
			  struct v_selection *sel, struct obj *dest,	\
			  enum xref_changed *changed, struct type *t)

#define T_IMPL_INIT(pfx, type, fd, ini, ini0, ini00, fini) \
enum s_status pfx##_init (struct type **t, struct xref_db *rdb,      	\
			  struct xref_bind *robj, const struct obj *obj,\
			  struct ref_owner *p, const char *fname, 	\
                          int line)					\
{									\
	return t_init ((struct t_type **) t, #type, rdb, robj, obj, 	\
		       p, sizeof (struct type), fd, 			\
		       pfx##_c_ini (ini), pfx##_c_ini (ini0),		\
		       pfx##_c_ini (ini00), pfx##_c_fini (fini),       	\
		       pfx##_cache_xkey, fname, line);			\
}									\

#define T_IMPL_NEW(pfx, type) \
enum s_status pfx##_new (struct type **t, struct xref_db *rdb,		\
			 struct ref_owner *p, const char *fname,        \
                         int line)					\
{									\
	return pfx##_init (t, rdb, NULL, NULL, p, fname, line);		\
}									\

#define T_IMPL_CROSS(pfx, type) \
enum s_status pfx##_cross (struct type **t, struct type *st, 		\
			   struct xref_bind *robj, struct ob_bridge *ob,\
 			   struct ref_owner *p, const char *fname, 	\
			   int line)					\
{								        \
	struct type *nt;						\
	struct obj sobj, obj;						\
	enum s_status res;						\
									\
	if (st == NULL) {						\
		*t = NULL;						\
		return S_CSC_OK;					\
	}								\
									\
	if (st -> t.rdb == ob -> drdb) {				\
		pfx##_ref (st, p, fname, line);				\
		*t = st;						\
		return S_CSC_OK;				    	\
	}								\
									\
	res = pfx##_sync (NULL, robj, NULL, &sobj, NULL, st);		\
	if (res != S_CSC_OK)						\
		return res;						\
									\
	res = ob_lookup (ob, &sobj, &obj);				\
	obj_destroy (&sobj);						\
	if (res != S_CSC_OK)						\
		return res;						\
									\
	res = pfx##_init (&nt, ob -> drdb, robj, &obj, p, fname, line);	\
	obj_destroy (&obj);						\
	if (res != S_CSC_OK)						\
		return res;						\
									\
	*t = nt;							\
									\
	return S_CSC_OK;						\
}

#define T_IMPL_CACHE(pfx, type, fd, fini)				\
enum s_status pfx##_cache_init (struct xref_db *rdb)			\
{									\
	return t_cache_init (rdb, pfx##_cache_xkey, &pfx##_cache_closure);\
}								       	\
void pfx##_cache_destroy (struct xref_db *rdb)				\
									\
{									\
	t_cache_destroy (rdb, pfx##_cache_xkey);			\
}								       	\

#define T_IMPL_REF(pfx, type, fd) \
void pfx##_ref (struct type *t, struct ref_owner *p, 			\
                const char *fname, int line)				\
{									\
	t_ref (T_CAST_T (t), p, fname, line);				\
}									\

#define T_IMPL_UNREF(pfx, type, fd, fini) \
void pfx##_unref (struct type *t, struct ref_owner *p)			\
{									\
	t_unref (T_CAST_T (t), p, fd, pfx##_c_fini (fini));		\
}									\

#define T_IMPL_FUNREF(pfx, type) \
void pfx##_funref (struct ref_owner *p, struct xref_field *fd)		\
{									\
	pfx##_unref (*(struct type **) fd -> field, p);			\
}

#define T_IMPL_PARSER(pfx, type, ini, ini0, fini) \
enum s_status pfx##_parser (struct xref_db *rdb, struct xref_bind *robj,\
			    struct ref_owner *p, 			\
 			    struct obj *src, struct xref_field *fd,	\
			    const char *fname, int line)		\
{									\
	return pfx##_init (fd -> field, rdb, robj, src, p, fname, line);\
}									\

#define T_IMPL_SER(pfx, type) \
enum s_status pfx##_ser (struct xref_db *rdb, struct xref_bind *robj,	\
			 struct v_selection *sel, struct obj *obj, 	\
			 enum xref_changed *changed, struct xref_field *fd)\
{									\
	struct type **tp;						\
									\
	tp = fd -> field;						\
									\
	return pfx##_sync (rdb, robj, sel, obj, changed, *tp);		\
}									\

#define T_IMPL_SYNC(pfx, type, fd) \
enum s_status pfx##_sync (struct xref_db *rdb, struct xref_bind *robj,	\
			  struct v_selection *sel, struct obj *dest,	\
			  enum xref_changed *changed, struct type *t)	\
{									\
	struct t_type *tp;						\
									\
	tp = t != NULL ? T_CAST_T (t) : NULL;				\
									\
	return t_sync (rdb, robj, sel, dest, tp, changed, fd);		\
}

#define T_IMPLEMENTATIONS_S(pfx, type, fd, fini) \
const char *pfx##_cache_xkey = #pfx " cache bind key";			\
struct xref_pstable pfx##_pstable = { 					\
	sizeof (struct type *),						\
	pfx##_parser, pfx##_ser, pfx##_funref };		     	\
									\
struct t_cache_closure pfx##_cache_closure = {	 			\
	fd, (t_fini) fini						\
};

#define T_CAST(pfx, type) \
typedef enum s_status (*pfx##_ini) (struct type *, struct xref_db *,	\
				    struct xref_bind *, 		\
				    struct ref_owner *);		\
typedef void (*pfx##_fini) (struct type *,  struct ref_owner *);	\
									\
static t_ini pfx##_c_ini (pfx##_ini i)					\
{									\
	return (t_ini) i;						\
}									\
									\
static t_fini pfx##_c_fini (pfx##_fini f)				\
{									\
	return (t_fini) f;						\
}									\


#define T_IMPLEMENTATIONS_LOC0(pfx, type, fd, ini, ini0, ini00, fini) 	\
T_IMPLEMENTATIONS_S (pfx, type, fd, fini)				\
T_CAST (pfx, type)							\
T_IMPL_INIT (pfx, type, fd, ini, ini0, ini00, fini)			\
T_IMPL_CROSS (pfx, type)						\
T_IMPL_NEW (pfx, type)							\
T_IMPL_CACHE (pfx, type, fd, fini)					\
T_IMPL_REF (pfx, type, fd)						\
T_IMPL_UNREF (pfx, type, fd, fini)					\
T_IMPL_FUNREF (pfx, type)						\
T_IMPL_PARSER (pfx, type, ini, ini0, fini)				\
T_IMPL_SER (pfx, type)							\
T_IMPL_SYNC (pfx, type, fd)

#define T_IMPLEMENTATIONS_LOC(pfx, type, fd, ini, ini0, fini) \
T_IMPLEMENTATIONS_LOC0 (pfx, type, fd, ini, ini0, NULL, fini)

#define T_IMPLEMENTATIONS(pfx, type, fd) \
T_IMPLEMENTATIONS_LOC (pfx, type, fd, NULL, NULL, NULL)

/* Functions left out: init, unref, sync */
#define T_SUGAR_IMPLEMENTATIONS(pfx, type) \
T_IMPLEMENTATIONS_S (pfx, type, NULL, NULL)				\
T_IMPL_NEW (pfx, type)							\
T_IMPL_CROSS (pfx, type)						\
T_IMPL_REF (pfx, type, fd)						\
T_IMPL_FUNREF (pfx, type)						\
T_IMPL_PARSER (pfx, type, NULL, NULL, NULL)				\
T_IMPL_SER (pfx, type)

#define T_CALL(f, args...) f (args, __FILE__, __LINE__)

enum s_status t_init (struct t_type **t, const char *name,
		      struct xref_db *rdb, struct xref_bind *robj,
		      const struct obj *obj, 
		      struct ref_owner *p, size_t size, 
		      struct xref_field *fd,
		      t_ini ini, t_ini ini0, t_ini ini00, t_fini fini,
		      const char *bkey, const char *fname, int line);
void t_ref (struct t_type *t, struct ref_owner *p,
	    const char *fname, int line);
void t_unref (struct t_type *t, struct ref_owner *p, 
	      struct xref_field *fd, t_fini fini);
enum s_status t_sync (struct xref_db *rdb, struct xref_bind *robj,
		      struct v_selection *sel, struct obj *dest, 
		      struct t_type *t, enum xref_changed *changed, 
		      struct xref_field *fd);

enum s_status t_tag_dup (struct t_tag *stag, struct t_tag *dtag);
void t_tag_destroy (struct t_tag *tag);
int t_tag_is_valid (struct t_tag *tag);
int t_tag_same (const struct t_tag *tag1, const struct t_tag *tag2);

enum s_status t_tag_get_checksum (struct t_tag *tag, struct hash *cks);
enum s_status t_cache_init (struct xref_db *rdb, const char *xkey,
			    struct t_cache_closure *cache_closure);
void t_cache_destroy (struct xref_db *rdb, const char *xkey);
void t_cache_unref (struct t_type *type, struct ref_owner *p, 
		    struct t_cache_closure *closure);

#endif
