/*
 * PDFlow -- version.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
#ifndef version_h

#define version_h

#include "fmt/obj.h"

struct v_option {
	const char *name;
	int major, minor;
	int speed, bspeed, size, text;
};

struct v_optable {
	const char *name;
	struct v_option *options;
	void *values;
	size_t vsize;
};

struct v_pref {
	int major, minor;
	int speed, bspeed, size, text;
	pdf_bool_t skip_optimizer;
};

#define V_OPTION(name, maj, min, speed, bspeed, size, text) \
	{ name, maj, min, speed, bspeed, size, text }
#define V_OPTION_SPEED(name, maj, min, speed, size) \
	V_OPTION (name, maj, min, speed, 0, size, 1)
#define V_OPTION_BSPEED(name, maj, min, speed, bspeed, size) \
	V_OPTION (name, maj, min, speed, bspeed, size, 1)
#define V_END { NULL }

struct v_feature {
	struct v_feature *next;
	struct v_optable *table;
	void *value;
};

struct v_selection {
	int major, minor;
	struct v_pref *pref;
	struct v_feature *list;
};

enum s_status v_validate (struct obj_bstr *vspec, int v [2]);
const char *v_get_hdr (struct v_selection *sel);

enum s_status v_add_selection (struct v_selection *sel,
			       struct v_optable *table);
void v_destroy_selection (struct v_selection *sel);
int v_supported (struct v_pref *pref, int major, int minor);
void v_upgrade (struct v_selection *sel, int major, int minor);

void *v_get_fval (struct v_selection *sel, struct v_optable *table);

#endif
