/*
 * PDFlow -- pdfi.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "pdf/pdfi.h"
#include "fmt/pdf-syn.h"
#include "fmt/machine.h"
#include "graph/gstate.h"
#include "grid/area.h"

static enum s_status tr_q (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_Q (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_cm (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_w (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_J (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_j (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_M (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_d (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_ri (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_i (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_gs (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_m (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_l (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_c (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_v (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_y (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_h (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_re (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_sh (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_S (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_s (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_f (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_f_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_B (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_B_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_b (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_b_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_n (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);

static enum s_status tr_W (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_W_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_CS (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_cs (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_SC (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_SCN (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);
static enum s_status tr_sc (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_scn (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure);

static enum s_status tr_G (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_g (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_RG (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_rg (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_K (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);
static enum s_status tr_k (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure);

static enum s_status tr_Do (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_BI (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_ID (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_Tc (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_Tw (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_Tz (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_TL (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_Tf (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_Tr (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_Ts (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_BT (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_ET (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_Td (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_TD (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_Tm (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_T_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_Tj (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_quote (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure);
static enum s_status tr_dquote (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure);
static enum s_status tr_TJ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status tr_d0 (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);
static enum s_status tr_d1 (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure);

static enum s_status def_handler (struct m_stack *stack,
				  struct obj *sobj, int token,
				  void *closure);

static struct obj_keyword keywords [] = {
	{ "q", tr_q },
	{ "Q", tr_Q },
	{ "cm", tr_cm },
	{ "w", tr_w },
	{ "J", tr_J },
	{ "j", tr_j },
	{ "M", tr_M },
	{ "d", tr_d },
	{ "ri", tr_ri },
	{ "i", tr_i },
	{ "gs", tr_gs },

	{ "m", tr_m },
	{ "l", tr_l },
	{ "c", tr_c },
	{ "v", tr_v },
	{ "y", tr_y },
	{ "h", tr_h },
	{ "re", tr_re },
	{ "sh", tr_sh },

	{ "S", tr_S },
	{ "s", tr_s },
	{ "f", tr_f },
	{ "F", tr_f },
	{ "f*", tr_f_ },
	{ "B", tr_B },
	{ "B*", tr_B_ },
	{ "b", tr_b },
	{ "b*", tr_b_ },
	{ "n", tr_n },

	{ "W", tr_W },
	{ "W*", tr_W_ },

	{ "CS", tr_CS },
	{ "cs", tr_cs },
	{ "SC", tr_SC },
	{ "SCN", tr_SCN },
	{ "sc", tr_sc },
	{ "scn", tr_scn },
	{ "G", tr_G },
	{ "g", tr_g },
	{ "RG", tr_RG },
	{ "rg", tr_rg },
	{ "K", tr_K },
	{ "k", tr_k },

	{ "Do", tr_Do },

	{ "BI", tr_BI },
	{ "ID", tr_ID },

	{ "Tc", tr_Tc },
	{ "Tw", tr_Tw },
	{ "Tz", tr_Tz },
	{ "TL", tr_TL },
	{ "Tf", tr_Tf },
	{ "Tr", tr_Tr },
	{ "Ts", tr_Ts },

	{ "BT", tr_BT },
	{ "ET", tr_ET },
	{ "Td", tr_Td },
	{ "TD", tr_TD },
	{ "Tm", tr_Tm },
	{ "T*", tr_T_ },

	{ "Tj", tr_Tj },
	{ "'",  tr_quote },
	{ "\"", tr_dquote },
	{ "TJ", tr_TJ },

	{ "d0", tr_d0 },
	{ "d1", tr_d1 },

	{ NULL }
};

struct pdfi_interpreter {
	struct xref_db *rdb;
	struct ps_parser *parser;
};

struct pdfi_closure {
	struct pdfi_interpreter *interpreter;
	struct pl_state ps;
	struct gs_context *gc;
	struct gspg_page *page;
	struct rd_resdict *resdict;
	struct oc_properties *oc;

	int get_rect;
	pdf_real_t wx, wy;
	struct cobj_rect rect;
};


static enum s_status pc_error (struct pdfi_closure *pc, const char *fmt, ...) 
{
	enum s_status res;
	va_list ap;

	va_start (ap, fmt);
	res = pl_verror (&pc -> ps, fmt, ap);
	va_end (ap);

	return res;
}

static enum s_status typed_pop (struct m_stack *stack, void *closure,
				struct obj *obj, enum obj_type ot)
{
	enum s_status res;
	int tt;

	if (!m_pop (stack, obj, &tt))
		return pc_error (closure, "Stack underflow");
	
	if (!obj_cast (obj, ot)) {
		res = obj_error (obj, "Bad operand type");
		obj_destroy (obj);
		return res;
	}

	return S_CSC_OK;
}

static enum s_status tr_q (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;

	pc = closure;

	return gs_push (pc -> gc);
}

static enum s_status tr_Q (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	enum s_status res;
	int eos;

	pc = closure;
	res = gs_pop (pc -> gc, &eos);
	if (res != S_CSC_OK)
		return res;

	if (eos)
		return pc_error (pc, "GState stack underflow");

	return S_CSC_OK;
}

static enum s_status tr_cm (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct cobj_matrix coef;
	struct obj eobj;
	enum s_status res;
	int i, j;

	pc = closure;

	for (i = 2; i >= 0; i--) {
		coef.a [i][2] = i == 2 ? 1 : 0;
		for (j = 1; j >= 0; j--) {
			res = typed_pop (stack, closure, &eobj, OT_REAL);
			if (res != S_CSC_OK)
				return res;

			coef.a [i][j] = eobj.val.real;
		}
	}

	res = gs_set_ctm (pc -> gc, &coef);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_w (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct gs_graph_state *gss;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_graph (pc -> gc, GS_GO_OP, &gss);
	if (res != S_CSC_OK)
		return res;

	gss -> line_width = sobj.val.real;

	return S_CSC_OK;
}

static enum s_status tr_J (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct gs_graph_state *gss;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_INTEGER);
	if (res != S_CSC_OK)
		return res;

	res = gs_graph (pc -> gc, GS_GO_OP, &gss);
	if (res != S_CSC_OK)
		return res;

	gss -> line_cap = sobj.val.integer;

	return S_CSC_OK;
}

static enum s_status tr_j (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct gs_graph_state *gss;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_INTEGER);
	if (res != S_CSC_OK)
		return res;

	res = gs_graph (pc -> gc, GS_GO_OP, &gss);
	if (res != S_CSC_OK)
		return res;

	gss -> line_join = sobj.val.integer;

	return S_CSC_OK;
}

static enum s_status tr_M (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct gs_graph_state *gss;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_graph (pc -> gc, GS_GO_OP, &gss);
	if (res != S_CSC_OK)
		return res;

	gss -> miter_limit = sobj.val.real;

	return S_CSC_OK;
}

static enum s_status tr_d (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct obj array, phase;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &phase, OT_INTEGER);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &array, OT_ARRAY);
	if (res != S_CSC_OK)
		return res;

	res = gs_context_set_dash_pattern_alt (pc -> gc, phase.val.integer,
					       &array);
	obj_destroy (&array);
	if (res != S_CSC_OK)
		return res;

	

	return S_CSC_OK;
}

static enum s_status tr_ri (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_NAME);
	if (res != S_CSC_OK)
		return res;

	res = gs_context_set_rendering_intent (pc -> gc, &sobj);
	obj_destroy (&sobj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_i (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct gs_graph_state *gss;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_graph (pc -> gc, GS_GO_OP, &gss);
	if (res != S_CSC_OK)
		return res;

	gss -> flatness_tolerance = sobj.val.integer;

	return S_CSC_OK;
}

static enum s_status tr_gs (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct obj nobj, sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &nobj, OT_NAME);
	if (res != S_CSC_OK)
		return res;

	res = rd_get_ext_g_state (pc -> resdict, &nobj, &sobj);
	obj_destroy (&nobj);
	if (res != S_CSC_OK)
		return res;

	res = gs_context_set_dictionary (pc -> gc, &sobj, NULL);
	obj_destroy (&sobj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_m (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct obj xobj, yobj;
	enum s_status res;

	pc = closure;

	res = typed_pop (stack, closure, &yobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &xobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_move (pc -> gc, obj, xobj.val.real, yobj.val.real);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_l (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct obj xobj, yobj;
	enum s_status res;

	pc = closure;

	res = typed_pop (stack, closure, &yobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &xobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_lineto (pc -> gc, obj, xobj.val.real, yobj.val.real);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_c (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct obj xobj, yobj;
	struct obj cx1obj, cy1obj;
	struct obj cx2obj, cy2obj;
	enum s_status res;

	pc = closure;

	res = typed_pop (stack, closure, &yobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &xobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &cy2obj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &cx2obj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &cy1obj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &cx1obj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_bezierto (pc -> gc, obj, 
			   cx1obj.val.real, cy1obj.val.real,
			   cx2obj.val.real, cy2obj.val.real,
			   xobj.val.real, yobj.val.real);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_v (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct obj xobj, yobj;
	struct obj cx2obj, cy2obj;
	enum s_status res;

	pc = closure;

	res = typed_pop (stack, closure, &yobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &xobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &cy2obj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &cx2obj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_bezierto (pc -> gc, obj, 
			   pc -> gc -> curr_point.x, 
			   pc -> gc -> curr_point.y,
			   cx2obj.val.real, cy2obj.val.real,
			   xobj.val.real, yobj.val.real);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_y (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	struct obj xobj, yobj;
	struct obj cx1obj, cy1obj;
	enum s_status res;

	pc = closure;

	res = typed_pop (stack, closure, &yobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &xobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &cy1obj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &cx1obj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_bezierto (pc -> gc, obj, 
			   cx1obj.val.real, cy1obj.val.real,
			   xobj.val.real, yobj.val.real,
			   xobj.val.real, yobj.val.real);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_h (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_close (pc -> gc, obj);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_re (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct obj xobj, yobj;
	struct obj wobj, hobj;
	struct cobj_rect rect;
	enum s_status res;

	pc = closure;

	res = typed_pop (stack, closure, &hobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &wobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &yobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &xobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	rect.left = xobj.val.real;
	rect.top = yobj.val.real;
	rect.right = rect.left + wobj.val.real;
	rect.bottom = rect.top + hobj.val.real;

	res = gs_rect (pc -> gc, obj, &rect);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_sh (struct m_stack *stack,
			    struct obj *sobj, int token,
			    void *closure)
{
	struct gspg_stmt *stmt;
	struct pdfi_closure *pc;
	struct shd_shading *shd;
	struct obj obj;
	enum s_status res;

	pc = closure;

	res = typed_pop (stack, pc, &obj, OT_NAME);
	if (res != S_CSC_OK)
		return res;

	res = rd_get_shading (pc -> resdict, &obj, &shd,
			      ref_rown (&pc -> page -> refs));
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	res = gspg_stmt_add (pc -> page, pc -> gc, &stmt);
	if (res != S_CSC_OK) {
		shd_unref (shd, ref_rown (&pc -> page -> refs));
		return res;
	}

	stmt -> type = GSPG_ST_SHADING;
	stmt -> u.shading = shd;

	return S_CSC_OK;
}

static enum s_status paint_op (struct obj *obj, struct pdfi_closure *pc,
			       pdf_bool_t close,
			       pdf_bool_t stroke, pdf_bool_t fill,
			       enum gs_fill_rule rule)
{
	struct gs_paint paint;
	enum s_status res;

	if (close) {
		res = gs_close (pc -> gc, obj);
		if (res != S_CSC_OK)
			return res;
	}

	paint.stroke = stroke;
	paint.fill = fill;
	paint.rule = rule;

	res = gspg_stmt_add_path (pc -> page, pc -> gc, &paint, 0);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_S (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	return paint_op (obj, closure, 0, 1, 0, 0);
}

static enum s_status tr_s (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	return paint_op (obj, closure, 1, 1, 0, 0);
}

static enum s_status tr_f (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	return paint_op (obj, closure, 0, 0, 1, GS_FR_WINDING);
}

static enum s_status tr_f_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	return paint_op (obj, closure, 0, 0, 1, GS_FR_EVEN_ODD);
}

static enum s_status tr_B (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	return paint_op (obj, closure, 0, 1, 1, GS_FR_WINDING);
}

static enum s_status tr_B_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	return paint_op (obj, closure, 0, 1, 1, GS_FR_EVEN_ODD);
}

static enum s_status tr_b (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	return paint_op (obj, closure, 1, 1, 1, GS_FR_WINDING);
}

static enum s_status tr_b_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	return paint_op (obj, closure, 1, 1, 1, GS_FR_EVEN_ODD);
}

static enum s_status tr_n (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;

	pc = closure;

	return gs_path_reset (pc -> gc);
}

static enum s_status tr_W (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct pdfi_closure *pc;

	pc = closure;

	return gs_add_clipping_path (pc -> gc, pc -> gc -> path,
				     GS_FR_WINDING);

}

static enum s_status tr_W_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;

	pc = closure;

	return gs_add_clipping_path (pc -> gc, pc -> gc -> path,
				     GS_FR_EVEN_ODD);
}

static enum s_status get_colour_space (struct m_stack *stack,
				       struct pdfi_closure *pc,
				       struct gc_colour_state **cst)
{
	enum s_status res;
	struct obj obj;

	res = typed_pop (stack, pc, &obj, OT_NAME);
	if (res != S_CSC_OK)
		return res;

	res = gc_set_colour_state (&obj, pc -> interpreter -> rdb, 
				   pc -> resdict, cst,
				   gs_graph_owner (pc -> gc));
	obj_destroy (&obj);

	return res;
}

static enum s_status tr_CS (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	res = get_colour_space (stack, pc, &gss -> stroking_colour_state);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_cs (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;
	
	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	res = get_colour_space (stack, pc, &gss -> nonstroking_colour_state);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status set_colour (struct m_stack *stack,
				 struct pdfi_closure *pc,
				 struct gc_colour_state **cst)
{
	pdf_real_t data [gc_get_components (*cst)];
	struct ref_owner owner;
	struct pt_pattern *ptp;
	enum s_status res;
	struct obj name, obj;
	int i;

	ref_init_owner (&owner, "pdfi:set_colour");

	if (gc_is_pattern (*cst)) {
		res = typed_pop (stack, pc, &name, OT_NAME);
		if (res != S_CSC_OK)
			return res;

		res = rd_get_pattern (pc -> resdict, &name, &ptp, &owner);
		obj_destroy (&name);
		if (res != S_CSC_OK)
			return res;
	} else
		ptp = NULL;

	for (i = gc_get_components (*cst) - 1; i >= 0; i--) {
		res = typed_pop (stack, pc, &obj, OT_REAL);
		if (res != S_CSC_OK) {
			if (ptp != NULL)
				pt_unref (ptp, gs_graph_owner (pc -> gc));
			return res;
		}
		data [i] = obj.val.real;
	}

	res = gc_select_colour (cst, ptp, data, gs_graph_owner (pc -> gc));
	if (ptp != NULL)
		pt_unref (ptp, &owner);
	if (res != S_CSC_OK)
		return res;
		
	return S_CSC_OK;
}

static enum s_status tr_SC (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_colour (stack, closure, &gss -> stroking_colour_state);
}

static enum s_status tr_SCN (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_colour (stack, closure, &gss -> stroking_colour_state);
}

static enum s_status tr_sc (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_colour (stack, closure, &gss -> nonstroking_colour_state);
}

static enum s_status tr_scn (struct m_stack *stack,
			     struct obj *obj, int token,
			     void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_colour (stack, closure, &gss -> nonstroking_colour_state);
}

static enum s_status set_device_colour (struct m_stack *stack, 
					enum gc_colour_space_family family,
					struct obj *refobj,
					struct pdfi_closure *pc,
					struct gc_colour_state **cst)
{
	enum s_status res;

	res = gs_set_device_colour (pc -> gc, pc -> resdict,
				    cst, family);
	if (res != S_CSC_OK)
		return res;

	return set_colour (stack, pc, cst);
}

static enum s_status tr_G (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_device_colour (stack, GC_CSF_DEVICE_GRAY, obj, pc,
				  &gss -> stroking_colour_state);
}

static enum s_status tr_g (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_device_colour (stack, GC_CSF_DEVICE_GRAY, obj, pc,
				  &gss -> nonstroking_colour_state);
}

static enum s_status tr_RG (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;	

	return set_device_colour (stack, GC_CSF_DEVICE_RGB, obj, pc,
				  &gss -> stroking_colour_state);
}

static enum s_status tr_rg (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_device_colour (stack, GC_CSF_DEVICE_RGB, obj, pc,
				  &gss -> nonstroking_colour_state);
}

static enum s_status tr_K (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_device_colour (stack, GC_CSF_DEVICE_CMYK, obj, pc,
				  &gss -> stroking_colour_state);
}

static enum s_status tr_k (struct m_stack *stack,
			   struct obj *obj, int token,
			   void *closure)
{
	struct gs_graph_state *gss;
	struct pdfi_closure *pc;
	enum s_status res;

	pc = closure;

	res = gs_graph (pc -> gc, GS_GO_COLOUR, &gss);
	if (res != S_CSC_OK)
		return res;
	
	return set_device_colour (stack, GC_CSF_DEVICE_CMYK, obj, pc,
				  &gss -> nonstroking_colour_state);
}

static enum s_status tr_Do (struct m_stack *stack,
			    struct obj *sobj, int token,
			    void *closure)
{
	struct gspg_stmt *stmt;
	struct pdfi_closure *pc;
	struct xo_object *xo;
	struct obj obj;
	enum s_status res;

	pc = closure;

	res = typed_pop (stack, pc, &obj, OT_NAME);
	if (res != S_CSC_OK)
		return res;

	res = rd_get_xobject (pc -> resdict, &obj, 
			      pc -> oc, &xo, ref_rown (&pc -> page -> refs));
	obj_destroy (&obj);
	if (res != S_CSC_OK)
		return res;

	res = gspg_stmt_add (pc -> page, pc -> gc, &stmt);
	if (res != S_CSC_OK) {
		xo_unref (xo, ref_rown (&pc -> page -> refs));
		return res;
	}

	stmt -> type = GSPG_ST_XOBJECT;
	stmt -> u.xobject = xo;

	return S_CSC_OK;

}

static enum s_status create_inline_image (struct m_stack *stack,
					  struct pdfi_closure *pc,
					  struct obj *idobj,
					  struct obj *sobj)
{
	struct obj_dictionary dictionary;
	struct obj_dict_node dnode;
	struct obj_region region;
	const uint8_t *begin, *end;
	struct f_ptr ptr;
	struct obj key, val;
	enum s_status res;

	obj_init_dictionary (&dictionary);
	res = S_CSC_OK;
	while (1) {
		res = typed_pop (stack, pc, &val, OT_ANY);
		if (res != S_CSC_OK)
			break;

		if (val.type == OT_KEYWORD &&
		    !strcmp (val.val.keyword -> name, "BI")) {
			ptr = val.ptr;
			obj_destroy (&val);
			break;
		}

		res = typed_pop (stack, pc, &key, OT_NAME);
		if (res != S_CSC_OK) {
			obj_destroy (&val);
			break;
		}

		obj_init_dict_node (&dnode, &key, &val);
		res = obj_xfer_dict (&dictionary, &dnode);
		if (res != S_CSC_OK) {
			obj_destroy (&val);
			obj_destroy (&key);
			break;
		}
	}

	if (res != S_CSC_OK) {
		obj_dict_destroy (&dictionary);
		return res;
	}

	begin = idobj -> ptr.cptr + 3;

	res = pl_seek_bol (&pc -> ps, "EI");
	if (res != S_CSC_OK) {
		obj_dict_destroy (&dictionary);
		return res;
	}

	end = pc -> ps.ptr.cptr - 2;
	obj_init_file_region (&region, begin, end, pc -> ps.ptr.file);

	obj_set_stream (sobj, &ptr, &dictionary, &region);

	return S_CSC_OK;
}

static enum s_status create_xo_image (struct pdfi_closure *pc,
				      struct obj *sobj,
				      struct xo_object **xop)
{
	struct xref_bind robj;
	enum s_status res;

	xref_bind_init (&robj);
	xref_bind_data (&robj, rd_bkey, pc -> resdict);
	xref_bind_data (&robj, oc_properties_bkey, pc -> oc);

	res = T_CALL (xo_init, xop, pc -> resdict -> t.rdb, &robj,
		      sobj, ref_rown (&pc -> page -> refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_BI (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	return m_push (stack, obj, token);
}

static enum s_status tr_ID (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct gspg_stmt *stmt;
	struct pdfi_closure *pc;
	struct xo_object *xo;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	res = create_inline_image (stack, pc, obj, &sobj);
	if (res != S_CSC_OK)
		return res;
	
	res = gspg_expand_shorthands (&sobj.val.stream.dictionary);
	if (res != S_CSC_OK) {
		obj_destroy (&sobj);
		return res;
	}
	
	res = create_xo_image (pc, &sobj, &xo);
	obj_destroy (&sobj);
	if (res != S_CSC_OK)
		return res;

	res = gspg_stmt_add (pc -> page, pc -> gc, &stmt);
	if (res != S_CSC_OK) {
		T_CALL (xo_ref, xo, ref_rown (&pc -> page -> refs));
		return res;
	}

	stmt -> type = GSPG_ST_INLINE_IMAGE;
	stmt -> u.xobject = xo;

	return S_CSC_OK;
}

static enum s_status tr_Tc (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct gs_text_state *tws;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_text (pc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	tws -> char_space = sobj.val.real;

	return S_CSC_OK;
}

static enum s_status tr_Tw (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct gs_text_state *tws;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_text (pc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	tws -> word_space = sobj.val.real;

	return S_CSC_OK;
}

static enum s_status tr_Tz (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct gs_text_state *tws;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_text (pc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	tws -> scale = sobj.val.real;

	return S_CSC_OK;
}

static enum s_status tr_TL (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct gs_text_state *tws;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_text (pc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	tws -> leading = sobj.val.real;

	return S_CSC_OK;
}

static enum s_status tr_Tf (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct gf_font *font;
	struct obj sobj, fobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &fobj, OT_NAME);
	if (res != S_CSC_OK)
		return res;

	res = rd_get_font (pc -> resdict, &fobj, &font, 
			   gs_graph_owner (pc -> gc));
	obj_destroy (&fobj);
	if (res != S_CSC_OK)
		return res;

	res = gs_context_set_font (pc -> gc, font, sobj.val.real);
	gf_unref (font, gs_graph_owner (pc -> gc));
	if (res != S_CSC_OK)
		return res;


	return S_CSC_OK;
}

static enum s_status tr_Tr (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct gs_text_state *tws;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_INTEGER);
	if (res != S_CSC_OK)
		return res;

	res = gs_text (pc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	tws -> render = sobj.val.integer;

	return S_CSC_OK;
}

static enum s_status tr_Ts (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct gs_text_state *tws;
	struct obj sobj;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &sobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = gs_text (pc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;

	tws -> rise = sobj.val.real;

	return S_CSC_OK;
}


static enum s_status tr_BT (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	
	pc = closure;
	
	return gs_begin_text (pc -> gc, obj);
}

static enum s_status tr_ET (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	const struct gs_text_state *tws;
	struct gs_paint paint;
	enum s_status res;

	pc = closure;

	tws = gs_ro_text (pc -> gc);

	gs_render_to_paint (tws -> render, &paint);

	if (gs_t_clip (pc -> gc)) {
	    res = gs_add_clipping_path (pc -> gc, pc -> gc -> path,
					GS_FR_WINDING);
	    if (res != S_CSC_OK)
		    return res;
	}

	return paint_op (obj, pc, 0, paint.stroke, paint.fill, paint.rule);
}

static enum s_status tr_Td (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct obj tx, ty;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &ty, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &tx, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	gs_newline (pc -> gc, tx.val.real, ty.val.real);

	return S_CSC_OK;
}

static enum s_status tr_TD (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct gs_text_state *tws;
	struct obj tx, ty;
	enum s_status res;

	pc = closure;

	res = gs_text (pc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;
	
	res = typed_pop (stack, closure, &ty, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &tx, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	tws -> leading = -ty.val.real;

	gs_newline (pc -> gc, tx.val.real, ty.val.real);

	return S_CSC_OK;
}

static enum s_status tr_Tm (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct cobj_matrix m;
	struct obj cobj;
	enum s_status res;

	pc = closure;

	m.a [2][2] = 1;
	
	res = typed_pop (stack, closure, &cobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;
	m.a [2][1] = cobj.val.real;

	res = typed_pop (stack, closure, &cobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;
	m.a [2][0] = cobj.val.real;

	m.a [1][2] = 0;
	
	res = typed_pop (stack, closure, &cobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;
	m.a [1][1] = cobj.val.real;

	res = typed_pop (stack, closure, &cobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;
	m.a [1][0] = cobj.val.real;

	m.a [0][2] = 0;
	
	res = typed_pop (stack, closure, &cobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;
	m.a [0][1] = cobj.val.real;

	res = typed_pop (stack, closure, &cobj, OT_REAL);
	if (res != S_CSC_OK)
		return res;
	m.a [0][0] = cobj.val.real;

	res = gs_set_mat (pc -> gc, &m);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status tr_T_ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;

	pc = closure;

	gs_newline (pc -> gc, 0, -gs_ro_text (pc -> gc) -> leading);

	return S_CSC_OK;
}

static enum s_status tr_Tj (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct obj text;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &text, OT_STRING);
	if (res != S_CSC_OK)
		return res;

	res = gs_write (pc -> gc, &text);
	obj_destroy (&text);
	if (res != S_CSC_OK)
		return res;
		       
	return S_CSC_OK;
}

static enum s_status tr_quote (struct m_stack *stack,
			       struct obj *obj, int token,
			       void *closure)
{
	struct pdfi_closure *pc;
	struct obj text;
	enum s_status res;

	pc = closure;
	
	gs_newline (pc -> gc, 0, -gs_ro_text (pc -> gc) -> leading);

	res = typed_pop (stack, closure, &text, OT_STRING);
	if (res != S_CSC_OK)
		return res;

	res = gs_write (pc -> gc, &text);
	obj_destroy (&text);
	if (res != S_CSC_OK)
		return res;
		       
	return S_CSC_OK;
}

static enum s_status tr_dquote (struct m_stack *stack,
				struct obj *obj, int token,
				void *closure)
{
	struct pdfi_closure *pc;
	struct gs_text_state *tws;
	struct obj aw, ac, text;
	enum s_status res;

	pc = closure;
	
	gs_newline (pc -> gc, 0, -gs_ro_text (pc -> gc) -> leading);

	res = typed_pop (stack, closure, &text, OT_STRING);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &ac, OT_REAL);
	if (res != S_CSC_OK) {
		obj_destroy (&text);
		return res;
	}

	res = typed_pop (stack, closure, &aw, OT_REAL);
	if (res != S_CSC_OK) {
		obj_destroy (&text);
		return res;
	}

	res = gs_text (pc -> gc, GS_GO_OP, &tws);
	if (res != S_CSC_OK)
		return res;
	
	tws -> char_space = ac.val.real;
	tws -> word_space = aw.val.real;

	res = gs_write (pc -> gc, &text);
	obj_destroy (&text);
	if (res != S_CSC_OK)
		return res;
		       
	return S_CSC_OK;
}

static enum s_status tr_TJ (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct obj text;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &text, OT_ARRAY);
	if (res != S_CSC_OK)
		return res;

	res = gs_write (pc -> gc, &text);
	obj_destroy (&text);
	if (res != S_CSC_OK)
		return res;
		       
	return S_CSC_OK;
}

static enum s_status tr_d0 (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct obj wx, wy;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &wy, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &wx, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	pc -> wx = wx.val.real;
	pc -> wy = wy.val.real;

	if (pc -> get_rect)
		f_ptr_eof (&pc -> ps.ptr, NULL);
		       
	return S_CSC_OK;
}

static enum s_status tr_d1 (struct m_stack *stack,
			    struct obj *obj, int token,
			    void *closure)
{
	struct pdfi_closure *pc;
	struct obj urx, ury;
	struct obj llx, lly;
	struct obj wx, wy;
	enum s_status res;

	pc = closure;
	
	res = typed_pop (stack, closure, &ury, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &urx, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &lly, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &llx, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &wy, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	res = typed_pop (stack, closure, &wx, OT_REAL);
	if (res != S_CSC_OK)
		return res;

	pc -> wx = wx.val.real;
	pc -> wy = wy.val.real;

	pc -> rect.left = llx.val.real;
	pc -> rect.bottom = lly.val.real;

	pc -> rect.top = ury.val.real;
	pc -> rect.right = urx.val.real;

	if (pc -> get_rect)
		f_ptr_eof (&pc -> ps.ptr, NULL);
		       
	return S_CSC_OK;
}

static enum s_status def_handler (struct m_stack *stack,
				  struct obj *sobj, int token,
				  void *closure)
{
	struct obj obj;
	int tt;

	while (!m_empty (stack)) {
		m_pop (stack, &obj, &tt);
		obj_destroy (&obj);
	}

	obj_destroy (sobj);

	return S_CSC_OK;
}

struct feed_closure {
	int i;
	struct pg_contents *pgc;
};

static int feed (struct pl_state *ps, pdf_ofs_t *limit, void *closure)
{
	struct feed_closure *feedc;
	struct f_ptr ptr;

	feedc = closure;

	if (feedc -> i >= feedc -> pgc -> elements)
		return 0;

	f_ptr_init (&ptr, feedc -> pgc -> file [feedc -> i++], 0);
	pl_init_kw (ps, &ptr, ps -> keys);

	return 1;
}

enum s_status eval_t3_rect (struct rd_resdict *rd,
			    struct f_ptr *ptr,
			    struct gspg_page **output,
			    struct ref_owner *p,
			    struct cobj_rect *rect,
			    pdf_real_t *wx, pdf_real_t *wy)
{
	struct pdfi_interpreter *pi;
	struct pdfi_closure pc;
	enum s_status res;

	res = pdfi_create (rd -> t.rdb, &pi);
	if (res != S_CSC_OK)
		return res;

	res = gspg_page_new (&pc.page, NULL, 0, p);
	if (res != S_CSC_OK) {
		pdfi_destroy (pi);
		return res;
	}

	pc.oc = NULL;
	pc.interpreter = pi;
	pc.get_rect = 1;
	pc.wx = pc.wy = -1;
	cobj_rect_set_null (&pc.rect);
	res = T_CALL (gs_context_new, &pc.gc, pi -> rdb, p);
	if (res != S_CSC_OK) {
		gspg_page_unref (pc.page, p);
		pdfi_destroy (pi);
		return res;
	}

	res = S_CSC_OK;

	pl_init_kw (&pc.ps, ptr, keywords);	
	res = ps_eval (pi -> parser, def_handler, &pc.ps, 0, &pc, NULL, NULL);
	gs_context_unref (pc.gc, p);
	pdfi_destroy (pi);
	if (res != S_CSC_OK) {
		gspg_page_unref (pc.page, p);
		return res;
	}

	*output = pc.page;
	*rect = pc.rect;
	*wx = pc.wx;
	*wy = pc.wy;

	return S_CSC_OK;
}

enum s_status pdfi_get_area (struct rd_resdict *rd,
			     struct obj *stream,
			     struct cobj_rect *rect,
			     pdf_real_t *wx,
			     pdf_real_t *wy)
{
	struct ref_owner owner;
	struct f_file *file;
	struct gspg_page *page;
	struct f_ptr fptr;
	struct a_area *area;
	enum s_status res;

	ref_init_owner (&owner, "get_area");

	res = obj_stream_file (stream, &file);
	if (res != S_CSC_OK)
		return res;

	f_ptr_init (&fptr, file, 0);

	res = eval_t3_rect (rd, &fptr, &page, &owner, rect, wx, wy);
	f_unref (file);
	if (res != S_CSC_OK)
		return res;

	if (cobj_rect_is_null (rect)) {
		res = a_get_areas (&area, page, NULL, &owner);
		gspg_page_unref (page, &owner);
		if (res != S_CSC_OK)
			return res;

		*rect = area -> g.bounds;
		a_unref (area, &owner);
	} else
		gspg_page_unref (page, &owner);

	return S_CSC_OK;
}

enum s_status pdfi_eval (struct pdfi_interpreter *pi,
			 struct cat_catalog *cat,
			 struct pg_page *page,
			 struct gspg_page **output,
			 struct ref_owner *p)
{
	struct pdfi_closure pc;
	struct ref_owner owner;
	struct pg_contents pgc;
	struct f_ptr ptr;
	enum s_status res;
	struct feed_closure feedc;

	ref_init_owner (&owner, "pdf_interpreter");

	res = gspg_page_new (&pc.page, &page -> inh.media_box, 
			     pg_get_rotate (page), p);
	if (res != S_CSC_OK)
		return res;

	pc.resdict = page -> inh.resources;
	pc.oc = cat -> oc_properties;
	pc.wx = pc.wy = -1;
	cobj_rect_set_null (&pc.rect);
	pc.get_rect = 0;

	res = pg_page_get_contents (cat -> pages, page, &pgc);
	if (res != S_CSC_OK) {
		gspg_page_unref (pc.page, p);
		return res;
	}

	pc.interpreter = pi;
	res = T_CALL (gs_context_new, &pc.gc, pi -> rdb, &owner);
	if (res != S_CSC_OK) {
		pg_page_destroy_contents (&pgc);		
		gspg_page_unref (pc.page, p);
		return res;
	}

	res = S_CSC_OK;

	if (pgc.elements) {
		f_ptr_init (&ptr, pgc.file [0], 0);
		pl_init_kw (&pc.ps, &ptr, keywords);	
		feedc.pgc = &pgc;
		feedc.i = 1;

		res = ps_eval (pi -> parser, def_handler, &pc.ps, 
			       0, &pc, feed, &feedc);
	}
	if (res == S_CSC_OK)
		res = gs_done (pc.gc);

	pg_page_destroy_contents (&pgc);		
	gs_context_unref (pc.gc, &owner);
	
	if (res != S_CSC_OK) {
		gspg_page_unref (pc.page, p);
		return res;
	}

	*output = pc.page;

	return S_CSC_OK;
}

enum s_status pdfi_create (struct xref_db *rdb,
			   struct pdfi_interpreter **pi)
{
	struct pdfi_interpreter *p;
	enum s_status res;

	p = malloc (sizeof (struct pdfi_interpreter));
	if (p == NULL)
		return S_CSC_OUT_OF_MEMORY;

	p -> rdb = rdb;

	res = ps_create (rdb, &p -> parser);
	if (res != S_CSC_OK) {
		free (p);
		return res;
	}

	*pi = p;

	return S_CSC_OK;
}

void pdfi_destroy (struct pdfi_interpreter *pi)
{
	ps_destroy (pi -> parser);
	free (pi);
}


