/*
 * PDFlow -- pdfo.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <string.h>

#include "fmt/pdf-lex.h"
#include "graph/gstate.h"
#include "pdf/pdfw.h"

#define OPS_BEFORE_FLUSH 50

struct pdfo_optimizer {

	int ops;

	struct pdfw_writer *pdfw;

};

enum s_status pdfo_create (struct pdfw_writer *pdfw,
			   struct pdfo_optimizer **pdfop)
{
	struct pdfo_optimizer *pdfo;

	pdfo = malloc (sizeof (struct pdfo_optimizer));
	if (pdfo == NULL)
		return S_CSC_OUT_OF_MEMORY;

	pdfo -> ops = 0;
	pdfo -> pdfw = pdfw;

	*pdfop = pdfo;

	return S_CSC_OK;
}

void pdfo_destroy (struct pdfo_optimizer *pdfo)
{
	free (pdfo);
}

static enum s_status optimize (struct pdfo_optimizer *pdfo,
			       struct pdfw_opcode **head,
			       struct pdfw_opcode ***tail)
{
	return S_CSC_OK;
}

enum s_status pdfo_flush (struct pdfo_optimizer *pdfo, 
			  struct pdfw_opcode **head,
			  struct pdfw_opcode ***tail,
			  pdf_bool_t end)
{
	enum s_status res;

	if (++pdfo -> ops < OPS_BEFORE_FLUSH && !end)
		return S_CSC_OK;

	res = optimize (pdfo, head, tail);
	if (res != S_CSC_OK)
		return res;

	res = pdfw_ll_flush (pdfo -> pdfw);
	if (res != S_CSC_OK)
		return res;
		
	return S_CSC_OK;
}

