/*
 * PDFlow -- psi_private.h
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifndef psi_private_h

#define psi_private_h

#include "refs.h"
#include "ps/psi.h"
#include "fmt/machine.h"
#include "fmt/pdf-lex.h"

struct psi_dict_stack_node {

	struct psi_dict_stack_node *next;

	struct obj obj;
};

struct psi_ccmap {
	struct ref_owner owner;

	struct psi_ccmap *next;

	struct cmap_table *tables;

	struct obj dict;

	struct cmap *cmap;
};

struct psi_file_resource {
	struct psi_file_resource *next;
	char *name, *path;
};

struct psi_resources {
	int refs;
	struct psi_ccmap *ccmaps;
	struct psi_file_resource *cmapres;
};

struct psi_interpreter {
	struct m_machine *mach;
	struct m_state *state;
	struct psi_dict_stack_node *user_stack;
	struct obj_dictionary sys_dict;

	struct psi_resources *resources;
	struct psi_ccmap *ccmap;
};

struct psi_closure {
	struct psi_interpreter *interpreter;
	struct pl_state ps;
};

struct psi_procset {

	const char *name;

	struct obj_keyword *keywords;

	enum s_status (*init) (struct m_stack *stack, 
			       struct psi_closure *closure);
};

extern struct psi_procset psi_core_procset;
extern struct psi_procset psi_math_procset;

extern struct psi_procset psi_cmap_procset;

enum s_status psi_error (struct psi_closure *pc, const char *fmt, ...);
enum s_status psi_safe_pop (struct m_stack *stack, void *closure,
			    int nops, ...);
enum s_status psi_user_push (struct psi_interpreter *pi, struct obj *dict);
int psi_user_pop (struct psi_interpreter *pi);
int psi_user_top (struct psi_interpreter *pi, struct obj **obj);

enum s_status psi_push (struct m_stack *stack, struct obj *obj,
			int token);

enum s_status psi_add_procset (struct psi_interpreter *pi,
			       struct psi_procset *ps);

enum s_status psi_add_file_resource (struct psi_file_resource **list,
				     char *name, char *path);
void psi_destroy_file_resources (struct psi_file_resource **list);
struct psi_file_resource *psi_find_file_resource (struct psi_file_resource *l,
						  struct obj *name);

void psi_cid_destroy_cmap_res (struct psi_resources *res);
void psi_cid_destroy_cmap (struct psi_interpreter *pi);

enum s_status psi_cid_def_cmap (struct m_stack *stack,
				struct psi_closure *closure);
enum s_status psi_cid_add_cmap_dir (struct psi_resources *res,
				    const char *dir);
enum s_status psi_cid_get_cmap_by_name (struct psi_interpreter *pi,
					struct obj *name,
					struct ref_owner *p,
					struct cmap **cmap);
void psi_cid_get_last_cmap (struct psi_interpreter *pi,
			    struct ref_owner *p,
			    struct cmap **cmap);

#endif
