/*
 * PDFlow -- catalog.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "version.h"
#include "types/cobj.h"
#include "types/catalog.h"

static enum s_status ini0 (struct cat_catalog *cat, struct xref_db *rdb,
			   struct xref_bind *robj, struct ref_owner *p);

static struct cobj_enum_desc page_layout_desc [] = {
	{ CAT_PL_SINGLE_PAGE, "SinglePage" },
	{ CAT_PL_ONE_COLUMN, "OneColumn" },
	{ CAT_PL_TWO_COLUMN_LEFT, "TwoColumnLeft" },
	{ CAT_PL_TWO_COLUMN_RIGHT, "TwoColumnRight" },
	{ CAT_PL_TWO_PAGE_LEFT, "TwoPageLeft" },
	{ CAT_PL_TWO_PAGE_RIGHT, "TowPageRight" },

	{ CAT_PL_SINGLE_PAGE, "/Default" },
	{ 0, NULL }
};

static struct cobj_enum_desc page_mode_desc [] = {
	{ CAT_PM_USE_NONE, "UseNone" },
	{ CAT_PM_USE_OUTLINES, "UseOutlines" },
	{ CAT_PM_USE_THUMBS, "UseThumbs" },
	{ CAT_PM_FULL_SCREEN, "UseFullScreen" },
	{ CAT_PM_USE_OC, "UseOC" },
	{ CAT_PM_USE_ATTACHMENTS, "UseAttachments" },

	{ CAT_PM_USE_NONE, "/Default" },
	{ 0, NULL }
};

#define XREF_IMPLEMENTATION struct cat_catalog
#include "types/catalog.tab.h"

T_IMPLEMENTATIONS_LOC (cat, cat_catalog, &field, NULL, ini0, NULL)

static enum s_status ini0 (struct cat_catalog *cat, struct xref_db *rdb,
			   struct xref_bind *robj, struct ref_owner *p)
{
	enum s_status res;

	res = T_CALL (pg_pages_new,
		      &cat -> pages, rdb, ref_rown (&cat -> t.refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

/* Set version 1.4 so we can choose version when dump is over */
enum s_status cat_add_selection (struct v_selection *sel)
{
	if (v_supported (sel -> pref, 1, 4))
		v_upgrade (sel, 1, 4);

	return S_CSC_OK;
}

enum s_status cat_derive (struct cat_catalog *cat, struct cat_catalog *scat,
			  struct ob_bridge *ob)
{
	enum s_status res;

	res = ts_cross (&cat -> lang, &scat -> lang, ob);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}
