/*
 * PDFlow -- info.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "types/info.h"

static struct cobj_enum_desc trapped_desc [] = {

	{ INFO_T_TRUE, "True" },
	{ INFO_T_FALSE, "False" },
	{ INFO_T_UNKNOWN, "Unknown" },

	{ INFO_T_UNKNOWN, "/Default" },

	{ -1, NULL }

};

#define XREF_IMPLEMENTATION struct info
#include "types/info.tab.h"

T_IMPLEMENTATIONS (info, info, &field);

enum s_status info_touch (struct info *info)
{
	ts_destroy (&info -> producer);

	cobj_date_now (&info -> mod_date);

	return ts_init_c (&info -> producer, "PDFlow v." VERSION);
}

enum s_status info_update (struct info *info, struct hash_status *hs)
{
	enum s_status res;

	res = ts_update (&info -> title, hs);
	if (res != S_CSC_OK)
		return res;

	res = ts_update (&info -> author, hs);
	if (res != S_CSC_OK)
		return res;

	res = ts_update (&info -> subject, hs);
	if (res != S_CSC_OK)
		return res;

	res = ts_update (&info -> keywords, hs);
	if (res != S_CSC_OK)
		return res;

	res = ts_update (&info -> creator, hs);
	if (res != S_CSC_OK)
		return res;

	res = ts_update (&info -> producer, hs);
	if (res != S_CSC_OK)
		return res;

	res = hash_update (hs, &info -> creation_date.time, sizeof (time_t));
	if (res != S_CSC_OK)
		return res;

	res = hash_update (hs, &info -> mod_date.time, sizeof (time_t));
	if (res != S_CSC_OK)
		return res;

	res = hash_update (hs, &info -> trapped.type, 
			   sizeof (info -> trapped.type));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

