/*
 * PDFlow -- pattern.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "types/pattern.h"
#include "graph/gstate.h"

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop);


#define XREF_IMPLEMENTATION struct pt_pattern
#define PT_PATTERN
#include "types/pattern.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct pt_tiling_pattern
#define PT_TILING
#include "types/pattern.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct pt_shading_pattern
#define PT_SHADING
#include "types/pattern.tab.h"
#undef XREF_IMPLEMENTATION

T_IMPLEMENTATIONS (pt, pt_pattern, &pt_pattern_field);

struct pt_pattern_ifc {
	enum pt_pattern_type type;
	struct xref_field *fd;
};

static struct pt_pattern_ifc tiling_ifc = {
	PT_PT_TILING, &pt_tiling_field
};

static struct pt_pattern_ifc shading_ifc = {
	PT_PT_SHADING, &pt_shading_field
};

static struct pt_pattern_ifc *ifcs [] = {
	&tiling_ifc, &shading_ifc, NULL
};

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop)
{
	struct pt_pattern *pt;
	int i;

	pt = data;

	if (pt -> ifc == NULL) {
		for (i = 0; ifcs [i] != NULL; i++)
			if (pt -> pattern_type == ifcs [i] -> type)
				break;

		pt -> ifc = ifcs [i];
	}

	if (pt -> ifc == NULL) {
		switch (op) {
		case XREF_OP_PARSE:
			return obj_error (&pt -> t.obj, 
					  "Unsupported pattern type: %d",
					  pt -> pattern_type);
			break;

		case XREF_OP_SERIALIZE:
			return s_make_indirect 
				(S_CSC_INTERNAL_ERROR,
				 "Trying to serialize pattern type %lu",
				 pt -> pattern_type);

		case XREF_OP_DESTROY:
			/* Can only happen if no initialization
			 * has been done, otherwise the interface 
			 * should have already been set
			 */
			*stop = 1;
			return S_CSC_OK;
		}
	}

	*fd = *pt -> ifc -> fd;

	return S_CSC_OK;
}
