/*
 * PDFlow -- shading.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "types/shading.h"

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop);

#define XREF_IMPLEMENTATION struct shd_shading
#define SHD_SHADING
#include "types/shading.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct shd_s_function_based
#define SHD_FUNCTION_BASED
#include "types/shading.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct shd_s_axial
#define SHD_AXIAL
#include "types/shading.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct shd_s_radial
#define SHD_RADIAL
#include "types/shading.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct shd_s_free_gouraud
#define SHD_FREE_GOURAUD
#include "types/shading.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct shd_s_lattice_gouraud
#define SHD_LATTICE_GOURAUD
#include "types/shading.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct shd_s_coons_patch
#define SHD_COONS_PATCH
#include "types/shading.tab.h"
#undef XREF_IMPLEMENTATION

#define XREF_IMPLEMENTATION struct shd_s_tensor_patch
#define SHD_TENSOR_PATCH
#include "types/shading.tab.h"
#undef XREF_IMPLEMENTATION

T_IMPLEMENTATIONS (shd, shd_shading, &shd_shading_field);

struct shd_shading_ifc {

	enum shd_shading_type type;

	struct xref_field *fd;
};

static struct shd_shading_ifc shading_function_ifc = {
	SHD_ST_FUNCTION_BASED,
	&shd_function_based_field
};

static struct shd_shading_ifc axial_ifc = {
	SHD_ST_AXIAL,
	&shd_axial_field
};

static struct shd_shading_ifc radial_ifc = {
	SHD_ST_RADIAL,
	&shd_radial_field
};

static struct shd_shading_ifc free_gouraud_ifc = {
	SHD_ST_FREE_GOURAUD,
	&shd_free_gouraud_field
};

static struct shd_shading_ifc lattice_gouraud_ifc = {
	SHD_ST_LATTICE_GOURAUD,
	&shd_lattice_gouraud_field
};

static struct shd_shading_ifc coons_patch_ifc = {
	SHD_ST_COONS_PATCH,
	&shd_coons_patch_field
};

static struct shd_shading_ifc tensor_patch_ifc = {
	SHD_ST_TENSOR_PATCH,
	&shd_tensor_patch_field
};

static struct shd_shading_ifc *ifcs [] = {
	&shading_function_ifc, &axial_ifc, &radial_ifc, 
	&free_gouraud_ifc, &lattice_gouraud_ifc, 
	&coons_patch_ifc, &tensor_patch_ifc, NULL
};

static enum s_status dyn_ifc (enum xref_op op, struct xref_field *fd, 
			      void *data, int *stop)
{
	struct shd_shading *shd;
	int i;

	shd = data;

	if (shd -> ifc == NULL) {
		for (i = 0; ifcs [i] != NULL; i++)
			if (shd -> shading_type == ifcs [i] -> type)
				break;

		shd -> ifc = ifcs [i];
	}

	if (shd -> ifc == NULL) {
		switch (op) {
		case XREF_OP_PARSE:
			return obj_error (&shd -> t.obj, 
					  "Unsupported shading type: %d",
					  shd -> shading_type);
			break;

		case XREF_OP_SERIALIZE:
			return s_make_indirect 
				(S_CSC_INTERNAL_ERROR,
				 "Trying to serialize shading type %lu",
				 shd -> shading_type);

		case XREF_OP_DESTROY:
			/* Can only happen if no initialization
			 * has been done, otherwise the interface 
			 * should have already been set
			 */
			*stop = 1;
			return S_CSC_OK;
		}
	}

	*fd = *shd -> ifc -> fd;

	return S_CSC_OK;
}

