/*
 * PDFlow -- trailer.c
 *
 *  Copyright (c) 2010 Alberto Cuda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "hash.h"
#include "types/trailer.h"
#include "types/catalog.h"

#define XREF_IMPLEMENTATION struct tri_trailer
#include "types/trailer.tab.h"

static enum s_status ini0 (struct tri_trailer *tri, struct xref_db *rdb,
			   struct xref_bind *robj, struct ref_owner *p);

T_IMPLEMENTATIONS_LOC (tri, tri_trailer, &field, NULL, ini0, NULL);


static enum s_status ini0 (struct tri_trailer *tri, struct xref_db *rdb,
			   struct xref_bind *robj, struct ref_owner *p)
{
	enum s_status res;

	res = T_CALL (cat_new,
		      &tri -> root, rdb, ref_rown (&tri -> t.refs));
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status tri_derive (struct tri_trailer *tri, struct tri_trailer *stri,
			  struct ob_bridge *ob)
{
	enum s_status res;

	res = cobj_id_cross (&tri -> id, &stri -> id, ob);
	if (res != S_CSC_OK)
		return res;

	res = T_CALL (info_cross, &tri -> info, stri -> info, 
		      NULL, ob, ref_rown (&tri -> t.refs));
	if (res != S_CSC_OK)
		return res;

	res = cat_derive (tri -> root, stri -> root, ob);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

enum s_status tri_touch (struct tri_trailer *tri)
{
	enum s_status res;

	if (tri -> info == NULL) {
		res = T_CALL (info_new, &tri -> info, tri -> t.rdb, 
			      ref_rown (&tri -> t.refs));
		if (res != S_CSC_OK)	
			return res;
	}

	res = info_touch (tri -> info);
	if (res != S_CSC_OK)
		return res;

	return S_CSC_OK;
}

static enum s_status get_id (struct xref_db *rdb, struct tri_trailer *tri,
			     struct cobj_id *id)
{
	struct hash_status *hs;
	const char *fname;
	pdf_ofs_t size;
	struct hash hash;
	enum s_status res;

	res = hash_init (&hs);
	if (res != S_CSC_OK)
		return res;

	res = xref_get_file_info (rdb, &fname, &size);
	if (res != S_CSC_OK)
		return res;

	res = hash_update (hs, &tri -> info -> mod_date.time, sizeof (time_t));
	if (res != S_CSC_OK)
		return res;

	if (fname != NULL) {
		res = hash_update (hs, fname, strlen (fname));
		if (res != S_CSC_OK)
			return res;
	}

	res = hash_update (hs, &size, sizeof (size));
	if (res != S_CSC_OK)
		return res;

	res = info_update (tri -> info, hs);
	if (res != S_CSC_OK)
		return res;

	res = hash_final (hs, &hash);
	if (res != S_CSC_OK)
		return res;

	if (tri -> id.id1.s == NULL) {
		id -> id1.size = HASH_SIZE;
		id -> id1.s = malloc (HASH_SIZE);
		if (id -> id1.s == NULL)
			return S_CSC_OUT_OF_MEMORY;
		memcpy (id -> id1.s, &hash.data, HASH_SIZE);
	} else {
		res = obj_string_dup (&tri -> id.id1, &id -> id1);
		if (res != S_CSC_OK)
			return res;
	}

	id -> id2.size = HASH_SIZE;
	id -> id2.s = malloc (HASH_SIZE);
	if (id -> id2.s == NULL) {
		free (id -> id1.s);
		return S_CSC_OUT_OF_MEMORY;
	}
	memcpy (id -> id2.s, hash.data, HASH_SIZE);

	return S_CSC_OK;
}

enum s_status tri_fix_and_sync (struct xref_db *rdb, struct v_selection *sel,
				struct obj *obj, struct tri_trailer *tri)
{
	struct cobj_id old_id, id;
	enum s_status res;

	res = get_id (rdb, tri, &id);
	if (res != S_CSC_OK)
		return res;

	old_id = tri -> id;
	tri -> id = id;

	res = tri_sync (rdb, NULL, sel, obj, NULL, tri);
	if (res != S_CSC_OK) {
		tri -> id = old_id;
		obj_string_destroy (&id.id1);
		obj_string_destroy (&id.id2);
		return res;
	}

	obj_string_destroy (&old_id.id1);
	obj_string_destroy (&old_id.id2);

	return S_CSC_OK;
}
